
''' A module for representing triangulations along with laminations and mapping classes on them.

Provides one class: EquippedTriangulation. '''

from random import choice
from itertools import product
import re
import multiprocessing

import flipper

def inverse(word):
	''' Return the inverse of a word by reversing and swapcasing it. '''
	
	return tuple([letter.swapcase() for letter in reversed(word)])

def generate_ordering(letters):
	''' Return a function which determines if v >= w (with respect to the short-lex ordering).
	
	If v or w contains any letter not in letters then returns False. '''
	
	positions = dict([(letter, index) for index, letter in enumerate(letters)])
	return lambda v, w: all(x in positions for x in v) and \
		all(y in positions for y in v) and \
		[len(v)] + [positions[x] for x in v] >= [len(w)] + [positions[y] for y in w]

##########################################################################
### A helper function  that can be pickled for multiprocessing.
def _worker_thread_word(Q, A):
	while True:
		data = Q.get()
		if data is None: break
		
		surface, length, prefix, options = data
		# We need to rebuild the ordering as this couldn't be passed through the pickle.
		options['order'] = generate_ordering(options['letters'])
		for output in surface._all_words_joined(length, prefix, **options):
			A.put(output)
	A.put(None)

def _worker_thread_mapping_class(Q, A):
	while True:
		data = Q.get()
		if data is None: break
		
		surface, length, prefix, options = data
		# We need to rebuild the ordering as this couldn't be passed through the pickle.
		options['order'] = generate_ordering(options['letters'])
		for output in surface._all_mapping_classes(length, prefix, **options):
			A.put(output)
	A.put(None)


class EquippedTriangulation(object):
	''' This represents a triangulation along with a collection of named laminations and mapping classes on it.
	
	Most importantly this object can construct a mapping class from a string descriptor.
	See self.mapping_class for additional information. '''
	def __init__(self, triangulation, laminations, mapping_classes):
		assert(isinstance(triangulation, flipper.kernel.Triangulation))
		assert(isinstance(laminations, (dict, list, tuple)))
		assert(isinstance(mapping_classes, (dict, list, tuple)))
		
		self.triangulation = triangulation
		if isinstance(laminations, dict):
			assert(all(isinstance(key, flipper.StringType) for key in laminations))
			assert(all(isinstance(laminations[key], flipper.kernel.Lamination) for key in laminations))
			assert(all(laminations[key].triangulation == self.triangulation for key in laminations))
			self.laminations = laminations
		else:
			assert(all(isinstance(lamination, flipper.kernel.Lamination) for lamination in laminations))
			assert(all(lamination.triangulation == self.triangulation for lamination in laminations))
			self.laminations = dict(list(flipper.kernel.utilities.name_objects(laminations)))
		
		if isinstance(mapping_classes, dict):
			assert(all(isinstance(key, flipper.StringType) for key in mapping_classes))
			assert(all(isinstance(mapping_classes[key], flipper.kernel.Encoding) for key in mapping_classes))
			assert(all(mapping_classes[key].source_triangulation == self.triangulation for key in mapping_classes))
			assert(all(mapping_classes[key].is_mapping_class() for key in mapping_classes))
			assert(all(key.swapcase() not in mapping_classes for key in mapping_classes))
			
			self.pos_mapping_classes = dict(mapping_classes)
			self.neg_mapping_classes = dict((name.swapcase(), self.pos_mapping_classes[name].inverse()) for name in self.pos_mapping_classes)
			self.mapping_classes = dict(list(self.pos_mapping_classes.items()) + list(self.neg_mapping_classes.items()))
		else:
			assert(all(isinstance(mapping_class, flipper.kernel.Encoding) for mapping_class in mapping_classes))
			assert(all(mapping_class.source_triangulation == self.triangulation for mapping_class in mapping_classes))
			assert(all(mapping_class.is_mapping_class() for mapping_class in mapping_classes))
			
			self.pos_mapping_classes = dict(list(flipper.kernel.utilities.name_objects(mapping_classes)))
			self.neg_mapping_classes = dict((name.swapcase(), self.pos_mapping_classes[name].inverse()) for name in self.pos_mapping_classes)
			self.mapping_classes = dict(list(self.pos_mapping_classes.items()) + list(self.neg_mapping_classes.items()))
		
		self.zeta = self.triangulation.zeta
	
	@classmethod
	def from_tuple(cls, objects):
		
		''' Create an EquippedTriangulation from a list of triangulations, laminations and mapping classes.
		
		Objects must be a non-empty list where each item is:
			1) an triangulation (at most one may be given),
			2) a Lamination,
			3) an Encoding,
			4) a pair (String, Lamination),
			5) a pair (String, Encoding).
		
		All laminations / mapping classes must be defined on the same triangulation
		We will automatically name items of type 2) and 3) sequentially:
			a, b, ..., z, aa, ab, ... . '''
		
		triangulation = None
		laminations, mapping_classes = {}, {}
		unnamed_laminations, unnamed_mapping_classes = [], []
		for item in objects:
			if isinstance(item, flipper.kernel.Triangulation):
				if triangulation is None:
					triangulation = item
				else:
					if item != triangulation:
						raise ValueError('Only one triangulation may be given.')
			elif isinstance(item, flipper.kernel.Lamination):
				unnamed_laminations.append(item)
			elif isinstance(item, flipper.kernel.Encoding):
				unnamed_mapping_classes.append(item)
			elif isinstance(item, (list, tuple)) and len(item) == 2:
				name, item2 = item
				if isinstance(name, flipper.StringType):
					if isinstance(item2, flipper.kernel.Lamination):
						if name not in laminations:
							laminations[name] = item2
						else:
							raise ValueError('Laminations with identical names.')
					elif isinstance(item2, flipper.kernel.Encoding):
						if name not in mapping_classes:
							mapping_classes[name] = item2
						else:
							raise ValueError('Encodings with identical names.')
					else:
						raise ValueError('Each item given must be a Lamination, Encoding, (String, Lamination) or (String, Encoding).')
				else:
					raise ValueError('Item must be named by a string.')
			else:
				raise ValueError('Each item given must be: Triangulation, Lamination, Encoding, (String, Lamination) or (String, Encoding).')
		
		for name, lamination in flipper.kernel.utilities.name_objects(unnamed_laminations, laminations):
			laminations[name] = lamination
		
		for name, encoding in flipper.kernel.utilities.name_objects(unnamed_mapping_classes, mapping_classes):
			mapping_classes[name] = encoding
		
		if triangulation is None:
			if len(laminations) > 0:
				triangulation = list(laminations.values())[0].triangulation
			elif len(mapping_classes) > 0:
				triangulation = list(mapping_classes.values())[0].source_triangulation
			else:
				raise ValueError('A triangulation, Lamination or Encoding must be given.')
		
		if any(laminations[name].triangulation != triangulation for name in laminations):
			raise ValueError('All laminations must be on the same triangulation.')
		if any(mapping_classes[name].source_triangulation != triangulation for name in mapping_classes):
			raise ValueError('All mapping classes must go from the same triangulation.')
		if any(mapping_classes[name].target_triangulation != triangulation for name in mapping_classes):
			raise ValueError('All mapping classes must go to the same triangulation.')
		
		return cls(triangulation, laminations, mapping_classes)
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		lam_keys = sorted(self.laminations.keys())
		pos_keys = sorted(self.pos_mapping_classes.keys())
		return 'Triangulation with laminations: %s and mapping classes: %s.' % (lam_keys, pos_keys)
	
	def random_word(self, length, positive=True, negative=True, letters=None):
		''' Return a random word of the required length.
		
		The letters to choose from can be specified or, alternatively, the set
		of positive, negative or all (default) mapping classes can be used by using the 
		flags postive and negative. '''
		
		assert(isinstance(length, flipper.IntegerType))
		
		if letters is None:
			pos_keys = sorted(self.pos_mapping_classes.keys())
			neg_keys = sorted(self.neg_mapping_classes.keys())
			
			if positive and negative:
				letters = pos_keys + neg_keys
			elif positive and not negative:
				letters = pos_keys
			elif not positive and negative:
				letters = neg_keys
			else:
				raise TypeError('At least one of positive and negative must be allowed.')
		
		return '.'.join(choice(letters) for _ in range(length))
	
	def generate_skip(self, length, letters=None):
		''' Return a dictionary whose keys are substrings that cannot appear in reduced words. '''
		
		letters = letters if letters is not None else sorted(self.mapping_classes, key=lambda x: (len(x), x.lower(), x.swapcase()))
		order = generate_ordering(letters)
		
		skip = dict()
		# Start by finding some common relations:
		# Trivial relations.
		for letter in letters:
			skip[(letter, letter.swapcase())] = tuple()
		# Commuting and braiding.
		for a, b in product(letters, repeat=2):
			A, B = a.swapcase(), b.swapcase()
			if A in letters and B in letters:
				for relator in [(a, b, A, B), (a, b, a, B, A, B)]:
					if self.mapping_class('.'.join(relator)).is_identity():
						j = len(relator) // 2
						for k in range(len(relator)):  # Cycling.
							ww = relator[k:] + relator[:k]
							if order(ww[:j], inverse(ww[j:])) and ww[:j] != inverse(ww[j:]):
								if all(ww[l:m] not in skip for m in range(j+1) for l in range(m)):
									skip[ww[:j]] = inverse(ww[j:])
		
		# Then do the actual search to the given relator_len.
		temp_options = {
			'group': True,
			'conjugacy': True,
			'bundle': False,
			'exact': True,
			'letters': letters,
			'order': order,
			'skip': skip,
			'prefilter': None,
			'filter': None
			}
		for i in range(1, length+1):
			relators = [word for word in self._all_words_unjoined(i, tuple(), **temp_options) if self.mapping_class('.'.join(word)).is_identity()]
			for j in range(i // 2, i+1):  # Slice length.
				for relator in relators:
					for k in range(i):  # Cycling.
						ww = relator[k:] + relator[:k]
						if order(ww[:j], inverse(ww[j:])) and ww[:j] != inverse(ww[j:]):
							if all(ww[l:m] not in skip for m in range(j+1) for l in range(m)):
								skip[ww[:j]] = inverse(ww[j:])
		
		return skip
	
	def _all_words_unjoined(self, length, prefix, **options):
		''' Yield all words of given length.
		
		Users should not call directly but should use self.all_words(...) instead.
		Assumes that various options have been set. '''
		
		order = options['order']
		letters = options['letters']
		skip = options['skip']
		lp = len(prefix)
		lp2 = lp + 1
		
		if not options['exact'] or len(prefix) == length:
			prefix_inv = inverse(prefix)
			
			good = True
			if good and options['conjugacy'] and prefix[-1:] == inverse(prefix[:1]): good = False
			if good and options['conjugacy'] and not all(order(prefix[i:] + prefix[:i], prefix) for i in range(lp)): good = False
			if good and options['bundle'] and all(x in letters for x in prefix_inv):
				if not all(order(prefix_inv[i:] + prefix_inv[:i], prefix) for i in range(lp)): good = False
			if good and options['filter'] is not None and not options['filter'](prefix): good = False
			if good:
				yield prefix
		
		if len(prefix) < length:
			for letter in letters:
				prefix2 = prefix + (letter,)
				
				good = True
				if good and options['group'] and prefix and any(prefix2[i:] in skip for i in range(lp2)): good = False
				if good and options['conjugacy'] and not all(order(prefix2[i:2*i], prefix2[:min(i, lp2-i)]) for i in range(lp2 // 2, lp)): good = False
				if good and options['prefilter'] is not None and not options['prefilter'](prefix): good = False
				if good:
					for word in self._all_words_unjoined(length, prefix2, **options):
						yield word
		
		return
	
	def _all_words_joined(self, length, prefix, **options):
		for word in self._all_words_unjoined(length, prefix, **options):
			joined = '.'.join(word)
			yield joined if options['apply'] is None else options['apply'](joined)
		
		return
	
	def _all_mapping_classes(self, length, prefix, **options):
		for word in self._all_words_unjoined(length, prefix, **options):
			mapping_class = self.mapping_class('.'.join(word))
			yield mapping_class if options['apply'] is None else options['apply'](mapping_class)
		
		return
	
	def all_words(self, length, prefix=None, **options):
		''' Yield all words of at most the specified length.
		
		There are several equivalence relations defined on these words.
		Words may represent the same:
			- mapping class group element (==),
			- conjugacy class (~~), or
			- fibre class (~?).
		
		Valid options and their defaults:
			equivalence='bundle' -- equivalence relation to use. 'bundle', 'conjugacy', 'group', 'none'
			exact=False -- skip words that do not have exactly the required length.
			letters=self.mapping_classes - a list of available letters to use, in alphabetical order.
			skip=None -- an iterable containing substrings that cannot appear.
			relator_len=2 -- if skip is not given then search words of length at most this much looking for relations.
			prefilter=None -- filter the prefixes of words by this function.
			filter=None -- filter the words by this function.
			apply=None -- apply the given function to the words.
			cores=None -- how many cores to use.
			prefix_depth=4 -- depth to search for prefixes for other cores.
		
		Notes:
			- By default letters are sorted by (length, lower case, swapcase).
			- For the equivalence used bundle ==> conjugacy ==> group.
			- The function given toapply must be a pickleable if using multiple cores. '''
		
		# Put the prefix into standard form.
		if prefix is None:
			prefix = tuple()
		if isinstance(prefix, flipper.StringType):
			prefix = tuple(self.decompose_word(prefix))
		else:
			prefix = tuple(prefix)
		
		# Setup options:
		default_options = {
			'equivalence': 'bundle',
			'letters': sorted(self.mapping_classes, key=lambda x: (len(x), x.lower(), x.swapcase())),
			'exact': False,
			'skip': None,
			'relator_len': 2,  # 2 get equal generators, 4 gets commutators and 6 gets braids.
			'prefilter': None,
			'filter': None,
			'apply': None,
			'cores': None,
			'prefix_length': 4
			}
		
		# Install any missing options with defaults.
		for option in default_options:
			if option not in options: options[option] = default_options[option]
		
		# Set implications. Possible values for options['equivalence'] are:
		#  None, group, conjugacy, bundle
		options['bundle'] = options['equivalence'] == 'bundle'
		options['conjugacy'] = options['equivalence'] == 'conjugacy' or options['bundle']
		options['group'] = options['equivalence'] == 'group' or options['conjugacy']
		
		# Build the list of substrings that must be avoided.
		if options['skip'] is not None:
			options['skip'] = set(options['skip'])
		elif options['group']:
			options['skip'] = self.generate_skip(options['relator_len'], options['letters'])
		else:
			options['skip'] = set()
		
		# We need to save a copy of the options at this point to pass to the nodes (if we
		# are multiprocessing) as the order function we are about to build is anonymous
		# and so it can't be Pickled.
		node_options = dict(options)
		
		# Build the ordering based on the letters given.
		options['order'] = generate_ordering(options['letters'])
		
		if options['cores'] is None:
			# Just use the single core algorithm:
			for word in self._all_words_joined(length, prefix, **options):
				yield word
		else:
			temp_options = dict(options)
			temp_options['conjugacy'] = False
			temp_options['bundle'] = False
			temp_options['exact'] = True
			temp_options['filter'] = None
			if options['prefix_length'] <= length:
				prefixes = [(self, length, leaf, node_options) for leaf in self._all_words_unjoined(options['prefix_length'], prefix, **temp_options)]
			else:
				prefixes = []
			
			Q = multiprocessing.Queue()
			A = multiprocessing.Queue(maxlen=1000)
			P = [multiprocessing.Process(target=_worker_thread_word, args=(Q, A)) for i in range(options['cores'])]
			for p in P: p.daemon = True
			for p in P: p.start()
			
			Q.put((self, min(options['prefix_length']-1, length), prefix, node_options))
			for data in prefixes:
				Q.put(data)
			
			for i in range(options['cores']):
				Q.put(None)
			
			num_completed_cores = 0
			while num_completed_cores < options['cores']:
				result = A.get()
				if result is None:
					num_completed_cores += 1
				else:
					yield result
			
			for p in P: p.terminate()
	
	def all_mapping_classes(self, length, prefix=None, **options):
		''' Yield all mapping classes of at most the specified length.
		
		There are several equivalence relations defined on these mapping classes.
		We may try to only one of each:
			- mapping class group element (==),
			- conjugacy class (~~), or
			- fibre class (~?).
		
		Valid options and their defaults:
			equivalence='bundle' -- equivalence relation to use. 'bundle', 'conjugacy', 'group','none'
			exact=False -- skip words that do not have exactly the required length.
			letters=self.mapping_classes - a list of available letters to use, in alphabetical order.
			skip=None -- an iterable containing substrings that cannot appear.
			relator_len=2 -- if skip is not given then search words of length at most this much looking for relations.
			prefilter=None -- filter the prefixes of words by this function.
			filter=None -- filter the words by this function.
			apply=None -- apply the given function to the words.
			cores=None -- how many cores to use.
			prefix_depth=3 -- depth to search for prefixes for other cores.
		
		Notes:
			- By default letters are sorted by (length, lower case, swapcase).
			- For the equivalence used bundle ==> conjugacy ==> group.
			- The function given toapply must be a pickleable if using multiple cores. '''
		
		# Put the prefix into standard form.
		if prefix is None:
			prefix = tuple()
		if isinstance(prefix, flipper.StringType):
			prefix = tuple(self.decompose_word(prefix))
		else:
			prefix = tuple(prefix)
		
		# Setup options:
		default_options = {
			'equivalence': 'bundle',
			'letters': sorted(self.mapping_classes, key=lambda x: (len(x), x.lower(), x.swapcase())),
			'exact': False,
			'skip': None,
			'relator_len': 2,  # 2 get equal generators, 4 gets commutators and 6 gets braids.
			'prefilter': None,
			'filter': None,
			'apply': None,
			'cores': None,
			'prefix_length': 3
			}
		
		# Install any missing options with defaults.
		for option in default_options:
			if option not in options: options[option] = default_options[option]
		
		# Set implications. Possible values for options['equivalence'] are:
		#  None, group, conjugacy, bundle
		options['bundle'] = options['equivalence'] == 'bundle'
		options['conjugacy'] = options['equivalence'] == 'conjugacy' or options['bundle']
		options['group'] = options['equivalence'] == 'group' or options['conjugacy']
		
		# Build the list of substrings that must be avoided.
		if options['skip'] is not None:
			options['skip'] = set(options['skip'])
		elif options['group']:
			options['skip'] = self.generate_skip(options['relator_len'], options['letters'])
		else:
			options['skip'] = set()
		
		# We need to save a copy of the options at this point to pass to the nodes (if we
		# are multiprocessing) as the order function we are about to build is anonymous
		# and so it can't be Pickled.
		node_options = dict(options)
		
		# Build the ordering based on the letters given.
		options['order'] = generate_ordering(options['letters'])
		
		if options['cores'] is None:
			# Just use the single core algorithm:
			for word in self._all_mapping_classes(length, prefix, **options):
				yield word
		else:
			temp_options = dict(options)
			temp_options['conjugacy'] = False
			temp_options['bundle'] = False
			temp_options['exact'] = True
			temp_options['filter'] = None
			if options['prefix_length'] <= length:
				prefixes = [(self, length, leaf, node_options) for leaf in self._all_words_unjoined(options['prefix_length'], prefix, **temp_options)]
			else:
				prefixes = []
			
			Q = multiprocessing.Queue()
			A = multiprocessing.Queue()
			P = [multiprocessing.Process(target=_worker_thread_mapping_class, args=(Q, A)) for i in range(options['cores'])]
			for p in P: p.daemon = True
			for p in P: p.start()
			
			Q.put((self, min(options['prefix_length']-1, length), prefix, node_options))
			for data in prefixes:
				Q.put(data)
			
			for i in range(options['cores']):
				Q.put(None)
			
			num_completed_cores = 0
			while num_completed_cores < options['cores']:
				result = A.get()
				if result is None:
					num_completed_cores += 1
				else:
					yield result
			
			for p in P: p.terminate()
	
	def decompose_word(self, word):
		''' Return a list of mapping_classes keys whose concatenation is word and the keys are chosen greedly.
		
		Raises a KeyError if the greedy decomposition fails. '''
		
		assert(isinstance(word, flipper.StringType))
		
		# By sorting the available keys, longest first, we ensure that any time we
		# get a match it is as long as possible.
		available_letters = sorted(self.mapping_classes, key=len, reverse=True)
		decomposition = []
		for subword in word.split('.'):
			while subword:
				for letter in available_letters:
					if subword.startswith(letter):
						decomposition.append(letter)
						subword = subword[len(letter):]
						break
				else:
					raise TypeError('After extracting %s, the remaining %s cannot be greedly decomposed as a concatination of self.mapping_classes.' % ('.'.join(decomposition), word))
		
		return decomposition
	
	def mapping_class(self, word):
		''' Return the mapping class corresponding to the given word or a random one of given length if given an integer.
		
		The given word is decomposed using self.decompose_word and the composition
		of the mapping classes involved is returned.
		
		Raises a TypeError if the word does not correspond to a mapping class. '''
		
		assert(isinstance(word, flipper.StringType) or isinstance(word, flipper.IntegerType))
		
		if isinstance(word, flipper.IntegerType):
			word = self.random_word(word)
		
		name = word  # Record the current word so we can use it later to name the mapping class.
		
		# Remove any whitespace.
		word = word.replace(' ', '')
		
		# Check for balanced parentheses.
		counter = 0
		for letter in word:
			if letter == '(': counter += 1
			if letter == ')': counter -= 1
			if counter < 0: raise TypeError('Unbalanced parentheses.')
		if counter != 0: raise TypeError('Unbalanced parentheses.')
		
		# Expand out parenthesis powers.
		# This can fail with a TypeError.
		old_word = None
		while word != old_word:  # While a change was made.
			old_word = word
			for subword, power in re.findall(r'(\([\w_\.]*\))\^(-?\d+)', word):
				decompose = self.decompose_word(subword[1:-1])
				int_power = int(power)
				if int_power > 0:
					replacement = '.'.join(decompose) * int_power
				else:
					replacement = '.'.join(letter.swapcase() for letter in decompose[::-1]) * abs(int_power)
				word = word.replace(subword + '^' + power, replacement)
		
		# Remove any remaining parenthesis, these do not have a power and are treated as ^1
		word = word.replace('(', '').replace(')', '')
		
		# Expand out powers without parenthesis. Here we use a greedy algorithm and take the
		# longest mapping class occuring before the power. Note that we only do one pass and so
		# only all pure powers to be expanded once, that is 'aBBB^2^3' is not recognised.
		available_letters = sorted(self.mapping_classes, key=len, reverse=True)
		for letter in available_letters:
			for subword, power in re.findall(r'(%s)\^(-?\d+)' % letter, word):
				int_power = int(power)
				word = word.replace(subword + '^' + power, (letter if int_power > 0 else letter.swapcase()) * abs(int_power))
		
		# This can fail with a TypeError.
		sequence = [item for letter in self.decompose_word(word) for item in self.mapping_classes[letter]]
		return flipper.kernel.Encoding(sequence, _cache={'name': name}) if len(sequence) > 0 else self.triangulation.id_encoding()
	
	def lamination(self, name):
		''' Return the lamination given by name. '''
		
		return self.laminations[name]





from __future__ import print_function

import flipper

def main(verbose=False):
	if verbose: print('Running matrix tests.')
	
	M = flipper.kernel.Matrix([[2, 1], [1, 1]])
	N = flipper.kernel.Matrix([[1, -1], [-1, 2]])
	M_inv = flipper.kernel.Matrix([[1, -1], [-1, 2]])
	# M + N = [[3, 0], [0, 3]]
	# M - N = [[1, 2], [2, -1]]
	
	tests = [
		M * N == flipper.kernel.id_matrix(2),
		M.characteristic_polynomial() == flipper.kernel.Polynomial([1, -3, 1]),
		M.determinant() == 1,
		M.kernel() == flipper.kernel.Matrix([]),
		(M + N).determinant() == 9,
		(M - N).determinant() == -5,
		(M**2)**3 == (M**3)**2,  # Check that powers are associative.
		M.inverse() == M_inv,
		M.characteristic_polynomial()(M) == flipper.kernel.zero_matrix(2),  # Check Cayley--Hamilton theorem.
		]
	
	if not all(tests):
		if verbose: print(tests)
		return False
	
	return True

if __name__ == '__main__':
	print(main(verbose=True))





from __future__ import print_function

import flipper
import sys

def main(surface, length, sample=100):
	S = flipper.load(surface)
	count = 0
	for i in range(sample):
		mapping_class = S.mapping_class(length)
		if mapping_class.is_pseudo_anosov(): count += 1
		print('\rLength: %d, Computed %d/%d - %0.1f%% pA' % (length, (i+1), sample, float(count) * 100 / (i+1)), end='')
		sys.stdout.flush()

if __name__ == '__main__':
	import argparse
	parser = argparse.ArgumentParser(description='Sample random words and compute the percentage that are pseudo-Anosov.')
	parser.add_argument('--surface', type=str, default='S_2_1', help='name of example surface to use')
	parser.add_argument('--length', type=int, default=10, help='length of words to generate')
	parser.add_argument('--sample', type=int, default=100, help='number of samples to take')
	parser.add_argument('--show', action='store_true', default=False, help='show the source code of this example and exit')
	args = parser.parse_args()
	
	if args.show:
		print(open(__file__, 'r').read())
	else:
		main(args.surface, args.length, args.sample)




from __future__ import print_function
from math import log10 as log
from itertools import product

import flipper

interval = flipper.kernel.Interval.from_string

def main(verbose=False):
	if verbose: print('Running interval tests.')
	
	w = interval('0.10')
	x = interval('10000.0')
	y = interval('1.14571')
	z = interval('1.00000')
	a = interval('-1.200000')
	b = interval('1.4142135623')
	
	# Check:
	#	acc(I + J) >= min(acc(I), acc(J)) - 1,
	#	acc(I * J) >= min(acc(I), acc(J)) - log(I.lower + J.lower + 1)
	#	acc(I / J) >= min(acc(I), acc(J)) - log+(J)  # If J > I.
	#	acc(x * I) >= acc(I) - log+(x)
	
	for I, J in product([w, x, y, z, a, b], repeat=2):
		m = min(I.accuracy, J.accuracy)
		if not ((I + J).accuracy >= m - 1): return False
		if not ((I * J).accuracy >= m - log(max(I.lower + J.lower + 1, 1))): return False
		# if not ((I / J).accuracy >= m - J.log_plus()): return False  # Should only do this test when J > I.
	
	return True

if __name__ == '__main__':
	print(main(verbose=True))





''' A module for a dummy interface. '''

import flipper
from math import log10 as log

def gram_schmidt(basis):
	''' Return an orthonormal basis for the subspace generated by basis. '''
	
	basis = [list(row) for row in basis]
	
	dot = flipper.kernel.dot
	for i in range(len(basis)):
		for j in range(i):
			a, b = dot(basis[i], basis[j]), dot(basis[j], basis[j])
			basis[i] = [b * x - a * y for x, y in zip(basis[i], basis[j])]
	return basis

def project(vector, basis):
	''' Return the projection of vector to the subspace generated by basis. '''
	
	dot = flipper.kernel.dot
	orthogonal_basis = gram_schmidt(basis)
	linear_combination = [dot(vector, row) / dot(row, row) for row in orthogonal_basis]
	return [sum(a * b[i] for a, b in zip(linear_combination, orthogonal_basis)) for i in range(len(vector))]

def directed_eigenvector(action_matrix, condition_matrix):
	''' An implementation of flipper.kernel.symboliccomputation.directed_eigenvector() using pure Python.
	
	See the docstring of the above function for further details. '''
	
	dot = flipper.kernel.dot
	# Getting the square free representative makes this faster.
	eigenvalues = [eigenvalue.irreducible_representation() for eigenvalue in action_matrix.characteristic_polynomial().square_free().real_roots() if eigenvalue > 1]
	
	for eigenvalue in sorted(eigenvalues, reverse=True):
		# We will calculate the eigenvector ourselves.
		N = flipper.kernel.NumberField(eigenvalue)
		kernel_basis = (action_matrix - N.lmbda).kernel()  # Sage is much better at this than us for large matrices.
		if len(kernel_basis) == 1:  # If rank(kernel) == 1.
			[eigenvector] = kernel_basis
			
			# We might need to flip the eigenvector if we have the inverse basis.
			if sum(eigenvector) < 0: eigenvector = [-x for x in eigenvector]
			
			if flipper.kernel.matrix.nonnegative(eigenvector) and condition_matrix.nonnegative_image(eigenvector):
				return N.lmbda, eigenvector
		else:
			M = (flipper.kernel.id_matrix(condition_matrix.width).join(condition_matrix)) * kernel_basis.transpose()
			try:
				linear_combination = M.find_vector_with_nonnegative_image()
				return N.lmbda, kernel_basis.transpose()(linear_combination)
			except flipper.AssumptionError:  # Eigenspace is disjoint from the cone.
				pass
	
	raise flipper.ComputationError('No interesting eigenvalues in cell.')





''' A module of useful, generic functions; including input and output formatting. '''

import flipper

import itertools
from string import ascii_lowercase, digits, ascii_letters, punctuation
from fractions import gcd as gcd2
from functools import reduce as freduce

VISIBLE_CHARACTERS = digits + ascii_letters + punctuation

def string_generator(n, skip=None):
	''' Return a list of n usable names, none of which are in skip. '''
	
	assert(isinstance(n, flipper.IntegerType))
	assert(skip is None or isinstance(skip, (list, tuple, dict, set)))
	
	skip = set() if skip is None else set(skip)
	if n < 1: return []
	
	alphabet = ascii_lowercase
	results = []
	i = 0
	while True:
		i += 1
		for letters in itertools.product(alphabet, repeat=i):
			word = ''.join(letters)
			if word not in skip:
				results.append(word)
			if len(results) >= n:
				return results

def name_objects(objects, skip=None):
	''' Return a list of pairs (name, object). '''
	
	assert(isinstance(objects, (list, tuple)))
	
	return zip(string_generator(len(objects), skip), objects)

###############################################################################

def product(iterable, start=1, left=True):
	''' Return the product of start (default 1) and an iterable of numbers. '''
	
	value = None
	for item in iterable:
		if value is None:
			value = item
		elif left:
			value = item * value
		else:
			value = value * item
	if value is None: value = start
	
	return value

def encode_binary(sequence):
	''' Return the given sequence of 0's and 1's as a string in base 64. '''
	
	step = 6  # 2**step <= len(VISABLE_CHARACTERS)
	return ''.join(VISIBLE_CHARACTERS[int(''.join(str(x) for x in sequence[i:i+step]), base=2)] for i in range(0, len(sequence), step))

def gcd(numbers):
	''' Return the gcd of a list of numbers. '''
	
	return freduce(gcd2, numbers)





# This directory provides several profile cases. Eventually there should be one
# for each module in the kernel. Each one provides a function main(verbose=False).

#from . import algebraicapproximation
#from . import algebraicnumber
from . import encoding
from . import equippedtriangulation
#from . import interval
from . import invariant_lamination
#from . import lamination
#from . import matrix
#from . import numberfield
#from . import permutation
#from . import polynomial
from . import symboliccomputation
from . import triangulation





from __future__ import print_function

import flipper

def main():
	S = flipper.load('S_1_2')
	word = 'aCBACBacbaccbAaAcAaBBcCcBBcCaBaaaABBabBcaBbCBCbaaa'
	
	h = S.mapping_class(word)
	print('Built the mapping class h := \'%s\'.' % word)
	
	print('h has order %s (where 0 == infinite).' % h.order())
	print('h is %s.' % h.nielsen_thurston_type())
	
	try:
		lamination = h.invariant_lamination()
		dilatation = h.dilatation()
		print('h leaves L := %s projectively invariant.' % lamination.projective_string())
		print('and dilates it by a factor of %s.' % dilatation)
	except flipper.AssumptionError:
		print('We cannot find a projectively invariant lamination for h as it is not pseudo-Anosov.')

if __name__ == '__main__':
	import argparse
	parser = argparse.ArgumentParser(description='Compute some properties of a mapping class.')
	parser.add_argument('--show', action='store_true', default=False, help='show the source code of this example and exit')
	args = parser.parse_args()
	
	if args.show:
		print(open(__file__, 'r').read())
	else:
		main()





''' The flipper kernel.

Some of the functions and methods have assumptions on them. We denote
in a functions docstring:
	Assumes that ...
		If the assumptions are met then this function is guaranteed to terminate correctly.
		If not then this function will either:
			terminate correctly, OR
			a flipper.AssumptionError will be raised.
	
	Assumes (and checks) that ...
		If the assumptions are met then this function is guaranteed to terminate correctly.
		If not then this a flipper.AssumptionError will be raised. '''

import flipper.kernel.interface
from . import algebraicapproximation
from . import bundle
from . import encoding
from . import error
from . import equippedtriangulation
from . import flatstructure
from . import interval
from . import lamination
from . import matrix
from . import moves
from . import numberfield
from . import permutation
from . import polynomial
from . import splittingsequence
from . import symboliccomputation
from . import triangulation
from . import triangulation3
from . import types
from . import utilities

# Set up shorter names for all of the different classes.
Vertex = triangulation.Vertex
Edge = triangulation.Edge
Triangle = triangulation.Triangle
Triangulation = triangulation.Triangulation
Corner = triangulation.Corner
AlgebraicApproximation = algebraicapproximation.AlgebraicApproximation
Bundle = bundle.Bundle
Encoding = encoding.Encoding
AbortError = error.AbortError
ApproximationError = error.ApproximationError
AssumptionError = error.AssumptionError
ComputationError = error.ComputationError
FatalError = error.FatalError
EquippedTriangulation = equippedtriangulation.EquippedTriangulation
Vector2 = flatstructure.Vector2
FlatStructure = flatstructure.FlatStructure
Interval = interval.Interval
Lamination = lamination.Lamination
Matrix = matrix.Matrix
Move = moves.Move
Isometry = moves.Isometry
EdgeFlip = moves.EdgeFlip
Spiral = moves.Spiral
LinearTransformation = moves.LinearTransformation
NumberField = numberfield.NumberField
NumberFieldElement = numberfield.NumberFieldElement
PolynomialRoot = polynomial.PolynomialRoot
Permutation = permutation.Permutation
Polynomial = polynomial.Polynomial
SplittingSequence = splittingsequence.SplittingSequence
SplittingSequences = splittingsequence.SplittingSequences
Tetrahedron = triangulation3.Tetrahedron
Triangulation3 = triangulation3.Triangulation3

id_matrix = matrix.id_matrix
zero_matrix = matrix.zero_matrix
dot = matrix.dot
height_int = numberfield.height_int
norm = triangulation.norm
StringType = types.StringType
IntegerType = types.IntegerType
FileType = types.FileType

# Functions that help with construction.
create_triangulation = Triangulation.from_tuple
triangulation_from_iso_sig = Triangulation.from_string
create_equipped_triangulation = EquippedTriangulation.from_tuple

product = utilities.product
gcd = utilities.gcd





from __future__ import print_function

import flipper

def main(verbose=False):
	if verbose: print('Running polynomial tests.')
	
	f = flipper.kernel.Polynomial([-2, 0, 1])  # f = -2 + x^2.
	g = flipper.kernel.Polynomial([0, 2])  # g = 2x = f'.
	h = flipper.kernel.Polynomial([-2, 2, 1])  # h = -2 + 2x + x^2 = f + g.
	p1 = flipper.kernel.Polynomial([1, -7, 19, -26, 19, -7, 1])
	p2 = flipper.kernel.Polynomial([2, -3, 1])  # 2 - 3x + x^2 = (x - 2) (x - 1).
	
	tests = [
		g == f.derivative(),
		h == f + g,
		len(f.real_roots()) == 2,
		len(p1.real_roots()) == 3,
		len(p2.real_roots()) == 2
		]
	
	return all(tests)

if __name__ == '__main__':
	print(main(verbose=True))




from __future__ import print_function

import flipper
try:
	import snappy
except ImportError:
	snappy = None

from time import time

def test(surface, word, target):
	M = snappy.Manifold(target)
	# M = snappy.twister.Surface(surface).bundle(word)  # These should be the same.
	N = snappy.Manifold(flipper.load(surface).mapping_class(word).bundle().snappy_string())
	for _ in range(100):
		try:
			if M.is_isometric_to(N):
				return True
		except RuntimeError:
			pass  # SnapPy couldn't decide if these are isometric or not.
		M.randomize()
		N.randomize()
	
	return False

def main():
	database = 'census_monodromies'  # We could also load('knot_monodromies').
	print('Building mapping tori for each monodromy in:')
	print('\t%s' % database)
	
	unmatched = []
	
	for surface, word, target in flipper.census(database):
		print('Buiding: %s over %s (target %s).' % (word, surface, target))
		start_time = time()
		if not test(surface, word, target):
			print('Could not match %s on %s' % (word, surface))
			unmatched.append((surface, word, target))
		print('\tComputed in %0.3fs' % (time() - start_time))
	
	if unmatched:
		print('Unable to match:')
		for unmatch in unmatched:
			print(unmatch)

if __name__ == '__main__':
	import argparse
	parser = argparse.ArgumentParser(description='Process each entry in a census')
	parser.add_argument('--show', action='store_true', default=False, help='show the source code of this example and exit')
	args = parser.parse_args()
	
	if args.show:
		print(open(__file__, 'r').read())
	else:
		if snappy is None:
			print('This example requires SnapPy.')
			print('Please install it and try again.')
		else:
			main()





# This directory provides several test cases. Eventually there should be one
# for each module in the kernel. Each one provides a function main(verbose=False)
# which returns True if and only if the test succeeds.

from . import algebraicapproximation
from . import encoding
from . import equippedtriangulation
from . import interval
from . import lamination
from . import lll
from . import matrix
from . import numberfield
from . import permutation
from . import polynomial
from . import triangulation
from . import triangulation3





# Provides several Tkinter widgets:
#	1) Meter - a progress bar, and
#	2) SplitButton - a button with additional menu of actions.
#	3) AnimatedCanvas - a canvas that periodically updates.

try:
	import Tkinter as TK
	DOWN_ARROW = unichr(9660)
except ImportError:  # Python 3.
	import tkinter as TK
	DOWN_ARROW = chr(9660)

import os
from base64 import b64encode

def extract_contents(paths, output_file):
	''' Convert a list of images into a single Python script. '''
	
	contents = ["\t'''" + b64encode(open(path, 'rb').read()) + "'''" for path in paths]
	open(output_file, 'w').write('\nframes_contents = [\n' + ',\n'.join(contents) + '\n\t]\n')

class SplitButton(TK.Frame):
	def __init__(self, parent, commands, **options):
		''' A SplitButton is a button with an additional
		drop down menu of other commands.
		
		It is initalised by passing a list of pairs of the form:
		(label, command). The first item in this list is
		used as the main function of the button and others are
		added to the drop down menu, this must be provided.
		
		Additional options for the underlying Frame can also be
		passed in. '''
		
		# Initalise.
		TK.Frame.__init__(self, parent, **options)
		
		# Setup controls.
		if len(commands) == 0:
			raise TypeError('At least one label and command must be provided.')
		
		# Build the widget.
		main_text, main_command = commands[0]
		self._main = TK.Button(self, text=main_text, command=main_command)
		self._more = TK.Menubutton(self, text=DOWN_ARROW, relief='raised')
		self._menu = TK.Menu(self._more, tearoff=0)
		self._more['menu'] = self._menu
		self._main.grid(column=0, row=0, sticky='nsew')
		self._more.grid(column=1, row=0, sticky='nsew', pady=1)
		
		# Add each command to the menu.
		for label, command in commands[1:]:
			self._menu.add_command(label=label, command=command)
		
		# Make it resizable with 75% of the size going to the button.
		self.columnconfigure(0, weight=3)
		self.columnconfigure(1, weight=1)
		self.rowconfigure(0, weight=1)

class Meter(TK.Frame):
	''' A simple progress bar widget for TK.
	
	INITIALIZATION OPTIONS:
	The widget subclasses a TK.Frame and adds:
	
		fillcolour -- the colour that is used to indicate the progress of the
					 corresponding process; default is "orchid1".
		value -- a float value between 0.0 and 1.0 (corresponding to 0% - 100%)
				 that represents the current status of the process; values higher
				 than 1.0 (lower than 0.0) are automagically set to 1.0 (0.0); default is 0.0 .
		text -- the text that is displayed inside the widget; if set to None the widget
				displays its value as percentage; if you don't want any text, use text="";
				default is None.
		font -- the font to use for the widget's text; the default is system specific.
		textcolour -- the colour to use for the widget's text; default is "black".
	
	WIDGET METHODS:
	All methods of a TK.Frame can be used; additionally there are two widget specific methods:
	
		get() -- returns a tuple of the form (value, text)
		set(value, text) -- updates the widget's value and the displayed text;
							if value is omitted it defaults to 0.0 , text defaults to None .
	'''
	def __init__(self, master, width=300, height=20, bg='white', fillcolour='midnight blue', value=0.0, text=None, font=None, textcolour='black', *args, **kw):
		
		TK.Frame.__init__(self, master, bg=bg, width=width, height=height, *args, **kw)
		self._value = value
		
		self._canv = TK.Canvas(self, bg=self['bg'], width=self['width'], height=self['height'], highlightthickness=0, relief='flat', bd=0)
		self._canv.pack(fill='both', expand=1)
		self._rect = self._canv.create_rectangle(0, 0, 0, self._canv.winfo_reqheight(), fill=fillcolour, width=0)
		self._text = self._canv.create_text(self._canv.winfo_reqwidth()/2, self._canv.winfo_reqheight()/2, text='', fill=textcolour)
		if font: self._canv.itemconfigure(self._text, font=font)
		
		self.set(value, text)
		self.bind('<Configure>', self._update_coords)
	
	def _update_coords(self, event):
		'''Updates the position of the text and rectangle inside the canvas when the size of
		the widget gets changed.'''
		# Looks like we have to call update_idletasks() twice to make sure to get the results we expect
		self._canv.update_idletasks()
		self._canv.coords(self._text, self._canv.winfo_width()/2, self._canv.winfo_height()/2)
		self._canv.coords(self._rect, 0, 0, self._canv.winfo_width()*self._value, self._canv.winfo_height())
		self._canv.update_idletasks()
	
	def get(self):
		return self._value, self._canv.itemcget(self._text, 'text')
	
	def set(self, value=0.0, text=None, lower_value=0.0):
		# Make the value failsafe:
		value = min(max(value, 0.0), 1.0)
		if self._canv.winfo_width() * abs(self._value - value) > 1 or text is not None:  # Only bother updating if there has been a noticeable change.
			self._value = value
			if text is None:
				text = '%0.1f%%' % (100 * value)  # If no text is specified use the default percentage string.
			textcolour = 'black' if value < 0.5 else 'white'
			self._canv.coords(self._rect, self._canv.winfo_width() * lower_value, 0, self._canv.winfo_width() * value, self._canv.winfo_height())
			self._canv.itemconfigure(self._text, text=text, fill=textcolour)
			self._canv.update_idletasks()

class AnimatedCanvas(TK.Canvas, object):
	''' A simple animated canvas widget for TK which regularly cycles through a sequence of frames.
	
	INITIALIZATION OPTIONS:
	The widget subclasses a TK.Canvas and adds:
		- frames -- a list of TK.PhotoImages to show.
		- frames_contensts -- a list of base64 encoded images to show.
		- frames_path -- a path to a directory containing frames to load.
		- delay -- the delay (in ms) between showing each frame.
		- start_animated -- start animating as soon as the widget is created.
	
	WIDGET METHODS:
	All methods of a TK.Canvas can be used; additionally:
		start() -- start animating the canvas.
		stop() -- stop animating the canvas. '''
	def __init__(self, parent, frames_path=None, frames_contents=None, frames=None, delay=85, start_animated=True, **options):
		self.parent = parent
		
		if frames_contents is not None:
			self._frames = [TK.PhotoImage(data=frame_contents) for frame_contents in frames_contents]
		elif frames_path is not None:
			paths = sorted(os.listdir(frames_path), key=lambda x: (len(x), x))
			self._frames = [TK.PhotoImage(file=os.path.join(frames_path, path)) for path in paths]
		elif frames is not None:
			self._frames = frames
		else:
			raise ValueError('frames_path, frames_contents or frames must be given.')
		
		w, h = self._frames[0].width(), self._frames[0].height()
		if 'width' not in options:
			options['width'] = w + 10
		if 'height' not in options:
			options['height'] = h + 10
		TK.Canvas.__init__(self, parent, **options)
		
		self._running = False
		self._delay = delay
		self._count = 0
		if start_animated: self.start()
	
	def start(self):
		if not self._running:
			self._running = True
			self.animate()
	
	def stop(self):
		self._running = False
	
	def animate(self):
		if self._running:
			self.create_image((5, 5), anchor='nw', image=self._frames[self._count])
			self._count = (self._count + 1) % len(self._frames)
			self.parent.after(self._delay, self.animate)





''' A module for representing laminations on Triangulations.

Provides one class: Lamination. '''

import flipper

import heapq
try:
	from Queue import Queue
except ImportError:
	from queue import Queue

HASH_DENOMINATOR = 30

class Lamination(object):
	''' This represents a lamination on an triangulation.
	
	It is given by a list of its geometric intersection numbers and a
	list of its algebraic intersection numbers with the (oriented) edges
	of underlying triangulation. Note that:
	     ^L
	     |
	-----|------> e
	     |
	has algebraic intersection +1.
	
	Users should use Triangulation.lamination() to create laminations with,
	when the lamination is a curve, its algebraic intersection computed
	automatically.
	
	If remove_peripheral is True then the Lamination is allowed to rescale
	its weights (by a factor of 2) in order to remove any peripheral
	components / satifsy the triangle inequalities. '''
	def __init__(self, triangulation, geometric, algebraic):
		assert(isinstance(triangulation, flipper.kernel.Triangulation))
		assert(isinstance(geometric, (list, tuple)))
		assert(isinstance(algebraic, (list, tuple)))
		# We should check that geometric / algebraic satisfies reasonable relations.
		
		self.triangulation = triangulation
		self.zeta = self.triangulation.zeta
		self.geometric = list(geometric)
		self.algebraic = list(algebraic)
		assert(len(self.geometric) == self.zeta)
		assert(len(self.algebraic) == self.zeta)
		
		self._cache = {}  # For caching hard to compute results.
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return str(self.geometric)
	
	def projective_string(self):
		''' Return a string describing this lamination in PML. '''
		
		w = float(self.weight())
		return str([float(x) / w for x in self])
	
	def __iter__(self):
		return iter(self.geometric)
	
	def __call__(self, item):
		''' Return the geometric measure assigned to item. '''
		if isinstance(item, flipper.IntegerType):
			return self.geometric[flipper.kernel.norm(item)]
		elif isinstance(item, flipper.kernel.Edge):
			return self.geometric[item.index]
		else:
			return NotImplemented
	
	def __getitem__(self, item):
		''' Return the algebraic measure assigned to item. '''
		if isinstance(item, flipper.IntegerType):
			normalised = flipper.kernel.norm(item)
			return self.algebraic[normalised] * (1 if normalised == item else -1)
		elif isinstance(item, flipper.kernel.Edge):
			return self.algebraic[item.index] * item.sign()
		else:
			return NotImplemented
	
	def __len__(self):
		return self.zeta
	
	def __eq__(self, other):
		return self.triangulation == other.triangulation and \
			all(v == w for v, w in zip(self.geometric, other.geometric)) and \
			all(v == w for v, w in zip(self.algebraic, other.algebraic))
	def __ne__(self, other):
		return not (self == other)
	
	def __hash__(self):
		# This should be done better.
		return hash(tuple(self.geometric))
	
	def __add__(self, other):
		if isinstance(other, Lamination):
			if other.triangulation != self.triangulation:
				raise ValueError('Laminations must be on the same triangulation to add them.')
			
			geometric = [x + y for x, y in zip(self, other)]
			algebraic = [x + y for x, y in zip(self.algebraic, other.algebraic)]
			return Lamination(self.triangulation, geometric, algebraic)
		else:
			if other == 0:  # So we can use sum.
				return self
			else:
				return NotImplemented
	def __radd__(self, other):
		return self + other
	
	def __mul__(self, other):
		geometric = [other * x for x in self]
		algebraic = [other * x for x in self.algebraic]
		return Lamination(self.triangulation, geometric, algebraic)
	def __rmul__(self, other):
		return self * other
	
	def is_empty(self):
		''' Return if this lamination is equal to the empty lamination. '''
		
		return not any(self)
	
	def isometries_to(self, other):
		''' Return a list of isometries taking this lamination to other. '''
		
		assert(isinstance(other, Lamination))
		
		# We used to just test other == isom.encode()(self) but this is much faster.
		# Should we check algebraic too?
		return [isom for isom in self.triangulation.isometries_to(other.triangulation) if all(other(isom.index_map[i]) == self(i) for i in self.triangulation.indices)]
	
	def self_isometries(self):
		''' Returns a list of isometries taking this lamination to itself. '''
		
		return self.isometries_to(self)
	
	def all_projective_isometries(self, other):
		''' Return a list of isometries taking this lamination projectively to other. '''
		
		assert(isinstance(other, Lamination))
		
		# This could be faster, we could build a smaller collection of isometries to begin with.
		return [isometry for isometry in self.triangulation.isometries_to(other.triangulation) if other.projectively_equal(isometry.encode()(self))]
	
	def projectively_equal(self, other):
		''' Return if this lamination is projectively equal to other.
		
		Other must be on the same Triangulation as this lamination. '''
		
		assert(isinstance(other, Lamination))
		assert(other.triangulation == self.triangulation)
		
		# We can't do division so we have to cross multiply.
		return self * other.weight() == other * self.weight()
	
	def projectivise(self):
		''' Return a canonically rescalled version of this lamination.
		
		This is currently pretty slow. '''
		
		# Start by dividing out by the weight.
		w = self.weight()
		companion_matrix = w.companion_matrix()
		N = w.number_field
		new_entries = [companion_matrix.solve(entry.linear_combination) for entry in self]
		# Now we need to divide out by the denominator:
		det = companion_matrix.determinant()
		# So find the gcd of all of the terms and divide by that.
		common = flipper.kernel.gcd([coeff for entry in new_entries for coeff in entry] + [det])
		new_new_entries = [N.element([coeff // common for coeff in entry]) for entry in new_entries]
		
		return self.triangulation.lamination(new_new_entries, remove_peripheral=False)
	
	def projective_hash(self, precision=HASH_DENOMINATOR):
		''' Return a hashable object that is invariant under isometries and rescaling. '''
		
		# Normalise so that it is invaiant under rescaling and sort to make it invariant under isometries.
		# We'll try to preserve as much of the structure as possible to try to reduce hash collisions.
		w = self.weight()
		wi = w.interval_approximation(2*precision).change_denominator(2*precision)
		# Do we need to worry about division by zero?
		L = [(entry.interval_approximation(2*precision).change_denominator(2*precision) / wi).change_denominator(precision).lower for entry in self]
		# Other (slow) ways of projectivising:
		# L = [entry // w for entry in self]
		# L = self.projectivise()
		
		# In this version we'll store the sorted, cyclically ordered, triangles.
		triples = [tuple([L[edge.index] for edge in triangle]) for triangle in self.triangulation]
		return tuple(sorted([min(triple[i:] + triple[:i] for i in range(len(triple))) for triple in triples]))
	
	def weight(self):
		''' Return the sum of the geometric intersection numbers of this lamination. '''
		
		return sum(self.geometric)
	
	def is_multicurve(self):
		''' Return if this lamination is a multicurve. '''
		
		if self == self.triangulation.empty_lamination(): return False
		
		# This isn't quite right. We should allow NumberFieldElements too.
		if not all(isinstance(entry, flipper.IntegerType) for entry in self): return False
		
		for corner_class in self.triangulation.corner_classes:
			for corner in corner_class:
				weights = [self(edge) for edge in corner.triangle]
				dual_weights_doubled = [weights[1] + weights[2] - weights[0],
					weights[2] + weights[0] - weights[1],
					weights[0] + weights[1] - weights[2]]
				if any(dual_weight % 2 != 0 for dual_weight in dual_weights_doubled): return False
		
		return True
	
	def conjugate_short(self):
		''' Return an encoding which maps this lamination to a lamination with as little weight as possible.
		
		This lamination must be a multicurve. '''
		
		# Repeatedly flip to reduce the weight of this lamination as much as possible.
		# Let [v_i] := f(self), where f is the encoding returned by this method.
		#
		# Self is a curve and each component of S - self has a puncture if and only if:
		#  v_i \in {0, 1} and all-bar-two v_i's are 0.
		#
		# Self is a curve and a component of S - self has no punctures if and only if:
		#  v_i \in {0, 2} and [v_i // 2] does not correspond to a (multi)curve.
		#
		# Otherwise, self is a multicurve.
		
		assert(self.is_multicurve())
		
		lamination = self
		best_conjugation = conjugation = lamination.triangulation.id_encoding()
		
		def weight_change(lamination, edge_index):
			''' Return how much the weight would change by if this flip was done. '''
			
			a, b, c, d = lamination.triangulation.square_about_edge(edge_index)
			return max(lamination(a) + lamination(c), lamination(b) + lamination(d)) - 2 * lamination(edge_index)
		
		time_since_last_weight_loss = 0
		old_weight = lamination.weight()
		possible_edges = set([i for i in lamination.triangulation.indices if lamination(i) > 0])
		# If we ever fail to make progress more than once then the curve is as short as it's going to get.
		while time_since_last_weight_loss < 2 and old_weight > 2:
			# Find the edge which decreases our weight the most.
			# If none exist then it doesn't matter which edge we flip, so long as it meets the curve.
			edge_index = min([i for i in possible_edges if lamination.triangulation.is_flippable(i)], key=lambda i: weight_change(lamination, i))
			
			forwards = lamination.triangulation.encode_flip(edge_index)
			conjugation = forwards * conjugation
			lamination = forwards(lamination)
			new_weight = lamination.weight()
			if lamination(edge_index) == 0:
				possible_edges.remove(edge_index)
			
			if new_weight < old_weight:
				time_since_last_weight_loss = 0
				old_weight = new_weight
				best_conjugation = conjugation
			else:
				time_since_last_weight_loss += 1
		
		return best_conjugation
	
	def is_curve(self):
		''' Return if this lamination is a curve. '''
		
		if not self.is_multicurve(): return False
		
		conjugation = self.conjugate_short()
		short_lamination = conjugation(self)
		
		if short_lamination.weight() == 2: return True
		
		# See the conditions on conjugate_short as to why this works.
		if any(weight not in [0, 2] for weight in short_lamination): return False
		
		# If we have a longer curve then all vertices must be on one side of it.
		# So if we collapse the edges with weight 0 we must end up with a
		# one vertex triangulation.
		triangulation = short_lamination.triangulation
		vertex_numbers = dict(zip(list(triangulation.vertices), range(len(triangulation.vertices))))
		for edge_index in triangulation.indices:
			if self(edge_index) == 0:
				c1, c2 = triangulation.vertices_of_edge(edge_index)
				a, b = vertex_numbers[c1], vertex_numbers[c2]
				if a != b:
					x, y = max(a, b), min(a, b)
					for c in vertex_numbers:
						if vertex_numbers[c] == x: vertex_numbers[c] = y
		
		# If any corner class is numbered > 0 then we don't have a one vertex triangulation.
		if any(vertex_numbers.values()): return False
		
		# So either we have a single curve or we have a multicurve with two parallel components.
		# We can test for the latter by checking 
		# We can test for the latter by seeing if the halved curve has the correct weight.
		#if triangulation.lamination([v // 2 for v in short_lamination]).weight() == short_lamination.weight() // 2:
		#	return False
		if all(ddual % 4 == 0 for corner in triangulation.corners for ddual in short_lamination.dual_weights_doubled(corner)):
			return False
		
		return True
	
	def is_twistable(self):
		''' Return if this lamination is a twistable curve. '''
		
		# This is based off of self.encode_twist(). See the documentation there as to why this works.
		if not self.is_curve(): return False
		
		conjugation = self.conjugate_short()
		short_lamination = conjugation(self)
		
		return short_lamination.weight() == 2
	
	def is_halftwistable(self):
		''' Return if this lamination is a half twistable curve. '''
		
		# This is based off of self.encode_halftwist(). See the documentation there as to why this works.
		
		conjugation = self.conjugate_short()
		short_lamination = conjugation(self)
		# We used to start with:
		#   if not self.is_twistable(): return False
		# But this wasted a lot of cycles repeating the calculation twice.
		if not short_lamination.weight() == 2:
			return False
		
		triangulation = short_lamination.triangulation
		
		e1, e2 = [edge_index for edge_index in short_lamination.triangulation.indices if short_lamination(edge_index) > 0]
		
		a, b, c, d = triangulation.square_about_edge(e1)
		if short_lamination(a) == 1 and short_lamination(c) == 1:
			e1, e2 = e2, e1
			a, b, c, d = triangulation.square_about_edge(e1)
		elif short_lamination(b) == 1 and short_lamination(d) == 1:
			pass
		
		_, _, z, w = triangulation.square_about_edge(a.label)
		_, _, x, y = triangulation.square_about_edge(c.label)
		
		return z == ~w or x == ~y
	
	def stratum(self):
		''' Return a dictionary mapping each vertex of the underlying triangulation to the number of stratum exiting it.
		
		This is the number of bipods incident to the vertex. '''
		
		triangulation = self.triangulation
		return dict((vertex, sum(1 for corner in triangulation.corner_class_of_vertex(vertex) if self.is_bipod(corner))) for vertex in triangulation.vertices)
	
	def dual_weights_doubled(self, corner):
		''' Return the list of dual weights (doubled) associated to a corner. '''
		weights = [self(edge_index) for edge_index in corner.indices]
		return [
			weights[1] + weights[2] - weights[0],
			weights[2] + weights[0] - weights[1],
			weights[0] + weights[1] - weights[2]
			]
	
	def is_nullpod(self, corner):
		''' Return if the lamination looks like a nullpod with respect to the given corner. '''
		
		dual_weights_doubled = self.dual_weights_doubled(corner)
		return all(weight == 0 for weight in dual_weights_doubled)
	def is_monopod(self, corner):
		''' Return if the lamination looks like a monopod with respect to the given corner. '''
		
		dual_weights_doubled = self.dual_weights_doubled(corner)
		return dual_weights_doubled[0] > 0 and dual_weights_doubled[1] == 0 and dual_weights_doubled[2] == 0
	def is_bipod(self, corner):
		''' Return if the lamination looks like a bipod with respect to the given corner. '''
		
		dual_weights_doubled = self.dual_weights_doubled(corner)
		return dual_weights_doubled[0] == 0 and dual_weights_doubled[1] > 0 and dual_weights_doubled[2] > 0
	def is_tripod(self, triangle):
		''' Return if the lamination looks like a tripod in this triangle. '''
		
		dual_weights_doubled = self.dual_weights_doubled(triangle)
		return all(weight > 0 for weight in dual_weights_doubled)
	
	def tripod_regions(self):
		''' Return a list of all triangles in which this lamination looks like a tripod. '''
		
		return [triangle for triangle in self.triangulation if self.is_tripod(triangle)]
	
	def puncture_tripods(self):
		''' Return the encoding corresponding to puncturing the tripods of this lamination. '''
		
		to_puncture = self.tripod_regions()
		geometric = 2 * flipper.kernel.id_matrix(self.zeta)
		algebraic = flipper.kernel.id_matrix(self.zeta)
		
		zeta = self.zeta
		triangulation = self.triangulation
		num_vertices = triangulation.num_vertices
		
		vertex_map = dict()
		for vertex in triangulation.vertices:
			vertex_map[vertex] = flipper.kernel.Vertex(vertex.label, filled=vertex.filled)
		
		edge_map = dict()
		# Far away edges should go to an exact copy of themselves.
		for edge in triangulation.positive_edges:
			edge_map[edge] = flipper.kernel.Edge(vertex_map[edge.source_vertex], vertex_map[edge.target_vertex], edge.label)
			edge_map[~edge] = ~edge_map[edge]
		
		triangles = []
		for triangle in triangulation:
			if triangle in to_puncture:
				a, b, c = triangle.labels
				A, B, C = triangle.indices
				x, y, z = [vertex_map[vertex] for vertex in triangle.vertices]
				w = flipper.kernel.Vertex(num_vertices, filled=True)
				num_vertices += 1
				p, q, r = [edge_map[edge] for edge in triangle.edges]
				s, t, u = [flipper.kernel.Edge(w, x, zeta), flipper.kernel.Edge(w, y, zeta+1), flipper.kernel.Edge(w, z, zeta+2)]
				triangles.append(flipper.kernel.Triangle([p, ~u, t]))
				triangles.append(flipper.kernel.Triangle([q, ~s, u]))
				triangles.append(flipper.kernel.Triangle([r, ~t, s]))
				
				# Note that we can't use zeta from above as that is tracking the *new* number of edges.
				geometric = geometric.join(flipper.kernel.zero_matrix(self.zeta, 1).tweak([(0, B), (0, C)], [(0, A)]))
				geometric = geometric.join(flipper.kernel.zero_matrix(self.zeta, 1).tweak([(0, C), (0, A)], [(0, B)]))
				geometric = geometric.join(flipper.kernel.zero_matrix(self.zeta, 1).tweak([(0, A), (0, B)], [(0, C)]))
				
				algebraic = algebraic.join(flipper.kernel.zero_matrix(self.zeta, 1))
				if c == C:
					algebraic = algebraic.join(flipper.kernel.zero_matrix(self.zeta, 1).tweak([(0, C)], []))
				else:
					algebraic = algebraic.join(flipper.kernel.zero_matrix(self.zeta, 1).tweak([], [(0, C)]))
				
				if b != B:
					algebraic = algebraic.join(flipper.kernel.zero_matrix(self.zeta, 1).tweak([(0, B)], []))
				else:
					algebraic = algebraic.join(flipper.kernel.zero_matrix(self.zeta, 1).tweak([], [(0, B)]))
				
				zeta += 3
			else:
				triangles.append(flipper.kernel.Triangle([edge_map[edge] for edge in triangle]))
		
		T = flipper.kernel.Triangulation(triangles)
		return flipper.kernel.LinearTransformation(triangulation, T, geometric, algebraic).encode()
	
	def collapse_trivial_weight(self, edge_index):
		''' Return this lamination on the triangulation obtained by collapsing edge edge_index
		and an encoding which is at least algebraically correct.
		
		Assumes (and checks) that:
			- edge_index is a flippable edge,
			- self.triangulation is not S_{0,3},
			- the given edge does not connect between two unfilled vertices, and
			- edge_index is the only edge of weight 0. '''
		
		if self(edge_index) != 0:
			raise flipper.AssumptionError('Lamination does not have weight 0 on edge to collapse.')
		
		# This relies on knowing how squares are returned.
		a, b, c, d = self.triangulation.square_about_edge(edge_index)  # Get the square about it.
		e = self.triangulation.edge_lookup[edge_index]
		
		# We'll first deal with some bad cases that con occur when some of the sides of the square are in fact the same.
		if a == ~b or c == ~d:
			# This implies that self(a) (respectively self(c)) == 0.
			raise flipper.AssumptionError('Additional weightless edge.')
		
		# There is at most one duplicated pair.
		if a == ~d and b == ~c:
			# This implies that the underlying surface is S_{0,3}.
			raise flipper.AssumptionError('Underlying surface is S_{0,3}.')
		
		if a == ~c and a == ~d:
			# This implies the underlying surface is S_{1,1}. As there is
			# only one vertex, both endpoints of this edge must be labelled 0.
			raise flipper.AssumptionError('Lamination is not filling.')
		
		# Now the only remaining possibilities are:
		#   a == ~c, b == ~d, a == ~d, b == ~c, or no relations.
		
		# Get the two vertices. We will ensure that if there is a filled vertex then bad_vertex is filled.
		bad_vertex, good_vertex = [e.source_vertex, e.target_vertex] if e.source_vertex.filled else [e.target_vertex, e.source_vertex]
		
		# We will quickly check some assumptions. If we collapse together two
		# unfilled vertices (or a vertex with itself) then there is a loop
		# disjoint to this lamination and so this is not filling.
		if (not good_vertex.filled and not bad_vertex.filled) or good_vertex == bad_vertex:
			raise flipper.AssumptionError('Lamination is not filling.')
		
		# Figure out how the vertices should be mapped.
		vertex_map = dict()
		# Every vertex, except the bad_vertex, goes to an exact copy of itself.
		for vertex in self.triangulation.vertices:
			if vertex != bad_vertex:
				# Although we will tweak the vertex labels slightly to ensure that they are still labelled 0, 1, ...
				new_label = vertex.label if vertex.label < bad_vertex.label else vertex.label - 1
				vertex_map[vertex] = flipper.kernel.Vertex(new_label, filled=vertex.filled)
		# The bad_vertex goes to the same place as the good_vertex.
		vertex_map[bad_vertex] = vertex_map[good_vertex]
		
		# Now figure out how the edges should be mapped.
		edge_count = 0
		edge_map = dict()
		# Far away edges should go to an exact copy of themselves.
		for edge in self.triangulation.positive_edges:
			if edge not in [a, b, c, d, e, ~a, ~b, ~c, ~d, ~e]:
				edge_map[edge] = flipper.kernel.Edge(vertex_map[edge.source_vertex], vertex_map[edge.target_vertex], edge_count)
				edge_map[~edge] = ~edge_map[edge]
				edge_count += 1
		
		if a == ~c:  # Collapsing an annulus.
			edge_map[~b] = flipper.kernel.Edge(vertex_map[b.target_vertex], vertex_map[b.source_vertex], edge_count)
			edge_map[~d] = ~edge_map[~b]
			edge_count += 1
		elif b == ~d:  # An annulus in the other direction.
			edge_map[~a] = flipper.kernel.Edge(vertex_map[a.target_vertex], vertex_map[a.source_vertex], edge_count)
			edge_map[~c] = ~edge_map[~a]
			edge_count += 1
		elif a == ~d:  # Collapsing a bigon.
			edge_map[~b] = flipper.kernel.Edge(vertex_map[b.target_vertex], vertex_map[b.source_vertex], edge_count)
			edge_map[~c] = ~edge_map[~b]
			edge_count += 1
		elif b == ~c:  # A bigon in the other direction.
			edge_map[~a] = flipper.kernel.Edge(vertex_map[a.target_vertex], vertex_map[a.source_vertex], edge_count)
			edge_map[~d] = ~edge_map[~a]
			edge_count += 1
		else:  # No identification.
			edge_map[~a] = flipper.kernel.Edge(vertex_map[a.target_vertex], vertex_map[a.source_vertex], edge_count)
			edge_map[~b] = ~edge_map[~a]
			edge_count += 1
			edge_map[~c] = flipper.kernel.Edge(vertex_map[c.target_vertex], vertex_map[c.source_vertex], edge_count)
			edge_map[~d] = ~edge_map[~c]
			edge_count += 1
		
		triples = [[edge_map[edge] for edge in triangle] for triangle in self.triangulation if e not in triangle and ~e not in triangle]
		new_triangles = [flipper.kernel.Triangle(triple) for triple in triples]
		T = flipper.kernel.Triangulation(new_triangles)
		
		bad_edges = [a, b, c, d, e, ~e]  # These are the edges for which edge_map is not defined.
		geometric = [[self(edge) for edge in self.triangulation.edges if edge not in bad_edges and edge_map[edge].index == i][0] for i in range(edge_count)]
		algebraic = [0] * edge_count
		lamination = Lamination(T, geometric, algebraic)
		
		# Now compute the encoding describing this. We will only bother to get the algebraic part correct
		# as the geometric part is not a PL function with only finitely many pieces.
		matrix = flipper.kernel.id_matrix(self.zeta)
		for edge in self.triangulation.edges:
			if edge != e and edge != ~e and edge.source_vertex == bad_vertex and edge.target_vertex != bad_vertex:
				matrix = matrix.elementary(edge.index, edge_index, +1 if edge.is_positive() != (e.source_vertex == bad_vertex) else -1)
		
		matrix2 = flipper.kernel.zero_matrix(self.zeta, edge_count)
		for edge in self.triangulation.edges:
			if edge not in bad_edges:
				target_edge = edge_map[edge]
				if not any(matrix2[target_edge.index]):
					matrix2[target_edge.index][edge.index] = +1 if edge.is_positive() == target_edge.is_positive() else -1
		
		algebraic_matrix = matrix2 * matrix
		geometric_matrix = flipper.kernel.zero_matrix(self.zeta, edge_count)
		
		encoding = flipper.kernel.LinearTransformation(self.triangulation, T, geometric_matrix, algebraic_matrix).encode()
		
		return lamination, encoding
	
	def splitting_sequences_uncached(self, dilatation=None, maxlen=None):
		''' Return a list of splitting sequence associated to this lamination.
		
		This is the encoding obtained by flipping edges to repeatedly split
		the branches of the corresponding train track with maximal weight
		until you reach a projectively periodic sequence (with the required
		dilatation if given).
		
		Assumes that this lamination is projectively invariant under some mapping class.
		Assumes (and checks) that this lamination is filling.
		
		Each entry of self.geometric must be an Integer or a NumberFieldElement (over
		the same NumberField). '''
		
		# In this method we use Lamination.projective_hash to store the laminations
		# we encounter efficiently and so avoid a quadratic algorithm. This currently
		# only ever uses the default precision HASH_DENOMINATOR. At some point this
		# should change dynamically depending on the algebraic numbers involved in
		# this lamination.
		
		assert(all(isinstance(entry, flipper.IntegerType) or isinstance(entry, flipper.kernel.NumberFieldElement) for entry in self))
		assert(len(set(entry.number_field for entry in self if isinstance(entry, flipper.kernel.NumberFieldElement))) <= 1)
		
		# Check if the lamination is obviously non-filling.
		if all(isinstance(entry, flipper.IntegerType) for entry in self):
			raise flipper.AssumptionError('Lamination is not filling.')
		
		# This assumes that no vertices are filled. We can probably get rid of this
		# if we get the collapse_trivial_weights preprocessing below working.
		if any(entry == 0 for entry in self):
			raise flipper.AssumptionError('Lamination is not filling.')
		
		# We should call lamination.collapse_trivial_weight(flip_index) on each
		# weight 0 edge and rely on it to either collapse the edge or raise the
		# approprate error.
		
		lamination = self
		encodings = []
		# Puncture all the triangles where the lamination is a tripod.
		E = lamination.puncture_tripods()
		lamination = E(lamination)
		encodings.append(E)
		
		# This is a dict taking the hash of each lamination to the index where we saw it.
		seen = {}
		# We then want a second dictionary taking indices where laminations occur back to
		# the lamination. This can use a lot of memory however as Tao's K(S) can grow very
		# large when the surface has high genus. To get around this we note that we will only
		# ever lookup laminations that occur in the last maxlen steps, which is at least the
		# periodic length of this sequence. So we create a bunch of dictionaries that we
		# cycle through. This way their union will always contain the last maxlen indices
		# and will only contain at most maxlen * NUM_LAM_BLOCKS / (NUM_LAM_BLOCKS - 1) laminations.
		NUM_LAM_BLOCKS = 5
		laminations = [dict() for i in range(NUM_LAM_BLOCKS)]
		current_block = 0  # This records which dictionary we are currently filling.
		
		# We don't include self or lamination in seen just incase they are already
		# on the axis and the puncture_tripods does nothing, misaligning the indices.
		# This ensures that at least one maximal split will be done.
		
		# We'll store the edge weights in a maximal heap using heapq. This allows us to quickly find the maximal weight edges.
		flip_first = lambda x: (-x[0], x[1])  # A small function allowing us to use Pythons defaul min heap as a max heap.
		weights_heap = [flip_first((weight, index)) for index, weight in enumerate(lamination)]
		heapq.heapify(weights_heap)
		# Get the index of the largest weight edge.
		flip_weight, flip_index = flip_first(heapq.heappop(weights_heap))
		while True:
			max_weight = flip_weight
			# Flip all edges weight max_weight. As the heap outputs these in sorted order,
			# we do this by popping an element and flipping it until we reach one of
			# weight < max_weight.
			while flip_weight == max_weight:
				# Do the flip.
				E = lamination.triangulation.encode_flip(flip_index)
				lamination = E(lamination)
				# Record information about the flip.
				encodings.append(E)
				
				# Check if we have created any edges of weight 0. Of course it is enough to just check flip_index.
				if lamination(flip_index) == 0:
					try:
						# If this fails it's because the lamination isn't filling.
						lamination, E = lamination.collapse_trivial_weight(flip_index)
						encodings.append(E)
						# Need to rebuild the heap as indices no longer correspond.
						weights_heap = [flip_first((weight, index)) for index, weight in enumerate(lamination)]
						heapq.heapify(weights_heap)
					except flipper.AssumptionError:
						raise flipper.AssumptionError('Lamination is not filling.')
				else:
					# Add the new edge weight back into the heap.
					heapq.heappush(weights_heap, flip_first((lamination(flip_index), flip_index)))
				
				# Get the next largest edge.
				flip_weight, flip_index = flip_first(heapq.heappop(weights_heap))
			
			# print(len(encodings))
			
			# Record this lamination in the dictionary of seen laminations.
			# To cut down on memory usage we will only retain the last maxlen laminations
			# at any given point.
			if maxlen is not None and len(laminations[current_block]) > maxlen // (NUM_LAM_BLOCKS - 1):
				# Move to the (cyclically) next block and reset it with just this lamination.
				current_block = (current_block + 1) % NUM_LAM_BLOCKS
				laminations[current_block] = {len(encodings): lamination}
			else:
				laminations[current_block][len(encodings)] = lamination
			
			# Check if lamination now (projectively) matches a lamination we've already seen.
			target = lamination.projective_hash()
			if target in seen:
				# print(seen[target])
				for index in seen[target]:
					for i in range(NUM_LAM_BLOCKS):
						if index in laminations[i]:
							old_lamination = laminations[i][index]
							break
					else:
						# This index has moved out of the lamination dictionaries and so is
						# too old to be the start point of the periodic cycle.
						continue
					
					# In the next block we have a lot of tests to do. We'll do these in
					# order of difficulty of computation. For example, computing
					# projective_isometries is slow; so we'll leave that to last to give
					# us the best chance that a faster test failing will allow us to
					# skip it.
					if dilatation is None or old_lamination.weight() >= dilatation * lamination.weight():
						isometries = lamination.all_projective_isometries(old_lamination)
						if len(isometries) > 0:
							assert(dilatation is None or old_lamination.weight() == dilatation * lamination.weight())
							# print('!!', index)
							
							encoding = flipper.kernel.Encoding([move for item in reversed(encodings) for move in item])
							return flipper.kernel.SplittingSequences(encoding, isometries, index, dilatation, old_lamination)
					else:
						# dilatation is not None and:
						#   old_lamination.weight() < dilatation * lamination.weight():
						# Note that the weight of laminations is strictly decreasing and the
						# indices of seen[target] are increasing. Thus if we are in this case
						# then the same inequality holds for every later index in seen[target].
						# Hence we may break out.
						break
				seen[target].append(len(encodings))
			else:
				# Start a new class containing this lamination.
				seen[target] = [len(encodings)]
	
	def splitting_sequences(self, dilatation=None, maxlen=None):
		''' A version of self.splitting_sequences_uncached with caching. '''
		
		if 'splitting_sequences' not in self._cache:
			self._cache['splitting_sequences'] = {}
		
		if dilatation not in self._cache['splitting_sequences']:
			try:
				self._cache['splitting_sequences'][dilatation] = self.splitting_sequences_uncached(dilatation, maxlen)
			except flipper.AssumptionError as error:
				self._cache['splitting_sequences'][dilatation] = error
		
		if isinstance(self._cache['splitting_sequences'][dilatation], Exception):
			raise self._cache['splitting_sequences'][dilatation]
		else:
			return self._cache['splitting_sequences'][dilatation]
	
	def encode_twist(self, k=1):
		''' Return an Encoding of a left Dehn twist about this lamination raised to the power k.
		
		This lamination must be a twistable curve. '''
		
		assert(self.is_twistable())
		
		if k == 0: return self.triangulation.id_encoding()
		
		conjugation = self.conjugate_short()
		short_lamination = conjugation(self)
		
		triangulation = short_lamination.triangulation
		# Grab the indices of the two edges we meet.
		e1, e2 = [edge_index for edge_index in short_lamination.triangulation.indices if short_lamination(edge_index) > 0]
		
		a, b, c, d = triangulation.square_about_edge(e1)
		# If the curve is going vertically through the square then ...
		if short_lamination(a) == 1 and short_lamination(c) == 1:
			# swap the labels round so it goes horizontally.
			e1, e2 = e2, e1
			a, b, c, d = triangulation.square_about_edge(e1)
		elif short_lamination(b) == 1 and short_lamination(d) == 1:
			pass
		
		# We now have:
		# #<----------#
		# |     a    ^^
		# |         / |
		# |---->------|
		# |       /   |
		# |b    e/   d|
		# |     /     |
		# |    /      |
		# |   /       |
		# |  /        |
		# | /         |
		# V/    c     |
		# #---------->#
		# And e.index = e1 and b.index = d.index = e2.
		
		# twist_k = flipper.kernel.Spiral(triangulation, triangulation, e1, k).encode()
		# return conjugation.inverse() * twist_k * conjugation
		
		twist = triangulation.encode([{i: i for i in triangulation.indices if i not in [e1, e2]}, e1])
		return conjugation.inverse() * twist**k * conjugation
	
	def encode_halftwist(self, k=1):
		''' Return an Encoding of a left half twist about this lamination raised to the power k.
		
		This lamination must be a half-twistable curve. '''
		
		assert(self.is_halftwistable())
		
		# This first section is the same as in self.encode_flip.
		if k == 0: return self.triangulation.id_encoding()
		
		conjugation = self.conjugate_short()
		short_lamination = conjugation(self)
		
		triangulation = short_lamination.triangulation
		e1, e2 = [edge_index for edge_index in short_lamination.triangulation.indices if short_lamination(edge_index) > 0]
		
		a, b, c, d = triangulation.square_about_edge(e1)
		# If the curve is going vertically through the square then ...
		if short_lamination(a) == 1 and short_lamination(c) == 1:
			# swap the labels round so it goes horizontally.
			e1, e2 = e2, e1
			a, b, c, d = triangulation.square_about_edge(e1)
		elif short_lamination(b) == 1 and short_lamination(d) == 1:
			pass
		
		# Get some more edges.
		_, _, z, w = triangulation.square_about_edge(a.label)
		_, _, x, y = triangulation.square_about_edge(c.label)
		
		# But now we have to go one further and worry about a, b, c, d Vs. c, d, a, b.
		# We want it so that x == ~y.
		if z.index == w.index:
			a, b, c, d = c, d, a, b
			w, x, y, z = y, z, w, x
		
		# So we now have:
		#       #
		#      / ^
		#     /   \
		#    /w   z\
		#   /       \
		#  V         \
		# #<----------#
		# |     a    ^^
		# |         / |
		# |---->------|
		# |       /   |
		# |b    e/   d|
		# |     /     |
		# |    /      |
		# |   /       |
		# |  /        |
		# | /         |
		# V/    c     |
		# #---------->#
		#  \         ^
		#   \       /
		#    \x   y/
		#     \   /
		#      V /
		#       #
		# Where e.index = e1 and b.index = d.index = e2,
		# and additionally x.index = y.index.
		
		half_twist = triangulation.encode([{i: i for i in triangulation.indices if i not in [e1, e2, c.index, x.index]}, e2, e1, c.index])
		
		# We accelerate large powers by replacting (T^1/2_self)^2 with T_self which includes acceleration.
		if abs(k) == 1:
			return conjugation.inverse() * half_twist**k * conjugation
		elif k % 2 == 0:  # k is even so use a Dehn twist
			return conjugation.inverse() * short_lamination.encode_twist(k // 2) * conjugation
		else:  # k is odd so we need to add in an additional half twist.
			# Note: k // 2 always rounds down, so even if k < 0 the additional half twist we need to do is positive.
			return conjugation.inverse() * short_lamination.encode_twist(k // 2) * half_twist * conjugation
	
	def geometric_intersection(self, lamination):
		''' Return the geometric intersection number between this lamination and the given one.
		
		Assumes (and checks) that this is a twistable lamination. '''
		
		assert(isinstance(lamination, Lamination))
		assert(lamination.triangulation == self.triangulation)
		
		if not self.is_twistable():
			raise flipper.AssumptionError('Can only compute geometric intersection number between a twistable curve and a lamination.')
		
		conjugator = self.conjugate_short()
		
		short = conjugator(self)
		short_lamination = conjugator(lamination)
		
		triangulation = short.triangulation
		e1, e2 = [edge_index for edge_index in triangulation.indices if short(edge_index) > 0]
		# We might need to swap these edge indices so we have a good frame of reference.
		if triangulation.corner_of_edge(e1).indices[2] != e2: e1, e2 = e2, e1
		
		a, b, c, d = triangulation.square_about_edge(e1)
		e = e1
		
		x = (short_lamination(a) + short_lamination(b) - short_lamination(e)) // 2
		y = (short_lamination(b) + short_lamination(e) - short_lamination(a)) // 2
		z = (short_lamination(e) + short_lamination(a) - short_lamination(b)) // 2
		x2 = (short_lamination(c) + short_lamination(d) - short_lamination(e)) // 2
		y2 = (short_lamination(d) + short_lamination(e) - short_lamination(c)) // 2
		z2 = (short_lamination(e) + short_lamination(c) - short_lamination(d)) // 2
		
		intersection_number = short_lamination(a) - 2 * min(x, y2, z)
		
		# Check that the other formula gives the same answer.
		assert(intersection_number == short_lamination(c) - 2 * min(x2, y, z2))
		
		return intersection_number
	
	def is_homologous_to(self, other, relative_boundary=False):
		''' Return if this lamination is homologous to the given one.
		
		The the homology class is computed relative to the filled punctures,
		unless relative_boundary is set to True in which case it is done
		relative to all vertices.
		
		This lamination and the given one must be defined on the same triangulation. '''
		
		assert(isinstance(other, Lamination))
		assert(self.triangulation == other.triangulation)
		
		matrix = flipper.kernel.id_matrix(self.zeta)
		
		triangulation = self.triangulation
		tree, dual_tree = triangulation.tree_and_dual_tree(not relative_boundary)
		vertices_used = dict((vertex, False) for vertex in triangulation.vertices)
		# Get some starting vertices.
		for vertex in triangulation.vertices:
			if not vertex.filled:
				vertices_used[vertex] = True
				if relative_boundary:  # Stop as soon as we've marked one.
					break
		
		while True:
			for edge in triangulation.edges:
				if tree[edge.index]:
					source, target = edge.source_vertex, edge.target_vertex
					if vertices_used[source] and not vertices_used[target]:  # This implies edge goes between distinct vertices.
						vertices_used[target] = True
						for edge2 in triangulation.edges:
							# We have to skip the edge2 == ~edge case at this point as we are still
							# removing it from various places.
							if edge2.source_vertex == target and edge2 != ~edge:
								matrix = matrix.elementary(edge2.index, edge.index, +1 if edge2.is_positive() == edge.is_positive() else -1)
						# Don't forget to go back and do edge which we skipped before.
						matrix = matrix.elementary(edge.index, edge.index, -1)
						break
			else:
				break  # If there are no more to add then we've dealt with every edge.
		
		M = flipper.kernel.Matrix([matrix[i] for i in range(self.zeta) if not tree[i] and not dual_tree[i]])
		return M(self.algebraic) == M(other.algebraic)
	
	def is_orientable(self):
		''' Return if the underlying train track of this lamination is orientable. '''
		
		if any(self.is_tripod(triangle) for triangle in self.triangulation):
			return False
		
		oris = [None] * self.zeta
		oris[0] = True
		to_process = Queue()
		to_process.put(0)
		to_process.put(~0)
		while not to_process.empty():
			to_do = to_process.get()
			corner = self.triangulation.corner_of_edge(to_do)
			rev_oris = [corner.indices[i] != corner.labels[i] for i in range(3)]
			dual_weights_doubled = self.dual_weights_doubled(corner)
			
			if dual_weights_doubled[1] > 0:
				new_ori = not oris[corner.indices[0]] ^ rev_oris[0] ^ rev_oris[2]
				if oris[corner.indices[2]] is None:
					oris[corner.indices[2]] = new_ori
					to_process.put(corner.labels[2])
					to_process.put(~corner.labels[2])
				elif oris[corner.indices[2]] != new_ori:
					return False
			if dual_weights_doubled[2] > 0:
				new_ori = not oris[corner.indices[0]] ^ rev_oris[0] ^ rev_oris[1]
				if oris[corner.indices[1]] is None:
					oris[corner.indices[1]] = new_ori
					to_process.put(corner.labels[1])
					to_process.put(~corner.labels[1])
				elif oris[corner.indices[1]] != new_ori:
					return False
		
		return True





''' A module for representing and manipulating elements of number fields.

Provides two classes: NumberField and NumberFieldElement. '''

# This provides us with a way of storing and manipulating elements of QQ(lambda),
# where lambda is an algebraic integer (however technically this can currently only actually
# manipulate elements of ZZ[lambda]). This can even do multiplication of these
# elements without having to drop to an AlgebraicApproximation and so is significantly
# faster that the previous way of doing this sort of calculation.
#
# This requires the numbers to be given as a linear combinations of
# 1, lambda, ..., lambda^{d-1}. Currently Sage is much better at doing this than we are.

import flipper

from math import log10 as log

LOG_2 = log(2)
INFTY = float('inf')

def height_int(number):
	''' Return the height of the given integer. '''
	
	assert(isinstance(number, flipper.IntegerType))
	
	return log(max(abs(number), 1))

class NumberField(object):
	''' This represents a number field QQ(lambda).
	
	Additionally NumberField() returns QQ.
	
	The given PolynomialRoot must be a monic polynomial. '''
	def __init__(self, polynomial_root=None):
		assert(polynomial_root is None or isinstance(polynomial_root, flipper.kernel.PolynomialRoot))
		if polynomial_root is None: polynomial_root = flipper.kernel.PolynomialRoot.from_tuple([-1, 1], '1.00')
		assert(polynomial_root.polynomial.is_monic())
		
		self.polynomial_root = polynomial_root
		
		self.height = self.polynomial_root.height
		self.degree = self.polynomial_root.degree
		self.log_degree = self.polynomial_root.log_degree
		self.accuracy_error = self.degree * self.polynomial_root.interval.log_plus()
		self.height_error = self.degree * (self.height + LOG_2)
		
		self.polynomial = self.polynomial_root.polynomial
		self.companion_matrices = self.polynomial.companion_matrix().powers(self.degree)
		
		self._approximations = None  # A list of approximations of lambda^0, ..., lambda^(d-1).
		self.accuracy = -1
		
		self.one = self.element([1])
		self.lmbda = self.element([1]) if self.is_rationals() else self.element([0, 1])  # lambda is a Python keyword.
	
	@classmethod
	def from_tuple(cls, coefficients, string):
		''' A short way of constructing a NumberField from a list of coefficients and a string. '''
		
		return cls(flipper.kernel.PolynomialRoot.from_tuple(coefficients, string))
	
	def lmbda_approximations(self, accuracy):
		''' Return intervals approximating lmbda^0, ..., lmbda^(degree-1) to the given accuracy.
		
		The intervals returned all have the same denominator and so addition / subtraction of these is fast. '''
		
		min_accuracy = 0
		target_accuracy = max(accuracy, min_accuracy)
		
		if self._approximations is None or self.accuracy < target_accuracy:
			# Note that if I is an interval approximating L to at
			# least k + i * N.log_plus digits of accuracy then I^i
			# approximates L^i to at least k digits of accuracy.
			request_accuracy = target_accuracy + self.accuracy_error
			
			lmbda = self.polynomial_root.interval_approximation(request_accuracy)
			self._approximations = [lmbda**i for i in range(self.degree)]
			minimal_accuracy = min(interval.accuracy for interval in self._approximations)
			if minimal_accuracy < INFTY:
				self._approximations = [interval.change_denominator(minimal_accuracy) for interval in self._approximations]
			self.accuracy = min(interval.accuracy for interval in self._approximations)
			assert(self.accuracy >= target_accuracy)
		
		return self._approximations
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return 'QQ[%s]' % str(self.polynomial)
	def __reduce__(self):
		# NumberFields are already pickleable but this results in a smaller pickle.
		return (self.__class__, (self.polynomial_root,))
	def __hash__(self):
		return hash(self.polynomial_root)
	def __eq__(self, other):
		return self.polynomial == other.polynomial and self.polynomial_root == other.polynomial_root
	def __ne__(self, other):
		return not (self == other)
	
	def element(self, linear_combination):
		''' Return a new element of this number field. '''
		
		return NumberFieldElement(self, linear_combination)
	
	def is_rationals(self):
		''' Return if this number field is QQ. '''
		
		return self.degree == 1

class NumberFieldElement(object):
	''' This represents an element of a number field. You shouldn't create NumberFieldElements directly but instead
	should use NumberField.element() which creates an element in that number field. '''
	def __init__(self, number_field, linear_combination):
		self.number_field = number_field
		if len(linear_combination) < self.number_field.degree:
			linear_combination = linear_combination + [0] * (self.number_field.degree - len(linear_combination))
		elif len(linear_combination) > self.number_field.degree:
			raise TypeError('Linear combination: %s has more terms than the degree of the field in which it lives' % linear_combination)
		self.linear_combination = linear_combination
		# Let N = QQ(lambda) and d := N.degree.
		# Let [a_i] := self.linear_combination, [\alpha_i] := N._algebraic_approximations[i].
		# Let \alpha := sum(a_i * \alpha_i).
		# Now h(\alpha) <= sum(h(a_i \alpha_i)) + d log(2)
		#               <= sum(h(a_i)) + sum(h(\alpha_i)) + (d-1) log(2) [AlgebraicApproximation.py L:9]
		#               <= sum(h(a_i)) + N.height_error
		self.degree = self.number_field.degree
		self.log_degree = self.number_field.log_degree
		self.height = sum(flipper.kernel.height_int(coefficient) for coefficient in self) + self.number_field.height_error
		self._interval = None
		self.accuracy = -1
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return self.approximate_string(6)
	def approximate_string(self, accuracy=None):
		''' Return a string approximating this NumberFieldElement. '''
		
		return self.algebraic_approximation(accuracy).approximate_string(accuracy)
	def __reduce__(self):
		# NumberFieldElements are already pickleable but this results in a smaller pickle.
		return (self.__class__, (self.number_field, self.linear_combination))
	def __iter__(self):
		return iter(self.linear_combination)
	def __len__(self):
		return self.number_field.degree
	def __float__(self):
		return float(self.algebraic_approximation())
	def __int__(self):
		return int(self.algebraic_approximation())
	def __bool__(self):
		return self != 0
	def __nonzero__(self):  # For Python2.
		return self.__bool__()
	def __hash__(self):
		# Warning: This is only really a hash when self.number_field.polynomial is irreducible.
		return hash((self.number_field, tuple(self.linear_combination)))
	
	def __neg__(self):
		return NumberFieldElement(self.number_field, [-a for a in self])
	def __abs__(self):
		return self if self >= 0 else -self
	def __add__(self, other):
		if isinstance(other, NumberFieldElement):
			if self.number_field != other.number_field:
				raise TypeError('Cannot add elements of different number fields.')
			return self.number_field.element([a+b for a, b in zip(self, other)])
		elif isinstance(other, flipper.IntegerType):
			return self.number_field.element([self.linear_combination[0] + other] + self.linear_combination[1:])
		else:
			return NotImplemented
	def __radd__(self, other):
		return self + other
	def __sub__(self, other):
		if isinstance(other, NumberFieldElement):
			if self.number_field != other.number_field:
				raise TypeError('Cannot subtract elements of different number fields.')
			return self.number_field.element([a-b for a, b in zip(self, other)])
		elif isinstance(other, flipper.IntegerType):
			return self.number_field.element([self.linear_combination[0] - other] + self.linear_combination[1:])
		else:
			return NotImplemented
	def __rsub__(self, other):
		return -(self - other)
	def companion_matrix(self):
		''' Return the companion matrix of this Element.
		
		This describes how multiplication by this Element acts on the underlying vector space. '''
		
		return flipper.kernel.dot(self, self.number_field.companion_matrices)
	def __mul__(self, other):
		if isinstance(other, NumberFieldElement):
			if self.number_field != other.number_field:
				raise TypeError('Cannot multiply elements of different number fields.')
			
			return self.number_field.element(self.companion_matrix()(other.linear_combination))
		elif isinstance(other, flipper.IntegerType):
			return self.number_field.element([a * other for a in self])
		else:
			return NotImplemented
	def __rmul__(self, other):
		return self * other
	def __pow__(self, power):
		if power == 0:
			return self.number_field.one
		elif power == 1:
			return self
		elif power > 1:
			sqrt = self**(power//2)
			square = sqrt * sqrt
			if power % 2 == 1:
				return self * square
			else:
				return square
	
	def __div__(self, other):
		return self.__truediv__(other)
	def __truediv__(self, other):
		if other == 0:
			raise ZeroDivisionError
		
		if isinstance(other, NumberFieldElement):
			if self.number_field != other.number_field:
				raise TypeError('Cannot multiply elements of different number fields.')
			
			I = flipper.kernel.id_matrix(self.degree+1)
			precision = int(self.height + other.height + self.degree) + 1
			while True:
				k = 10**precision
				
				lmbdas = self.number_field.lmbda_approximations(2*precision)
				
				a = self.interval_approximation(precision)
				b = other.interval_approximation(precision)
				
				M = I.join(flipper.kernel.Matrix([[int(k * lmbda) for lmbda in lmbdas] + [int(k * a / b)]])).transpose()
				N = M.LLL()  # This is really slow :(.
				div, scalar = self.number_field.element(N[0][:self.degree]), -N[0][-2]
				if div * other == scalar * self:
					return div, scalar
				else:
					# This should never happen if we chose precision correctly.
					# However Cohen described this choice as `subtle' so let's be
					# careful and repeat the calculation if we got the wrong answer.
					precision = 2 * precision
		elif isinstance(other, flipper.IntegerType):
			return self.number_field.element([coeff // other for coeff in self])
		else:
			return NotImplemented
	def __floordiv__(self, other):
		if other == 0:
			raise ZeroDivisionError
		
		if isinstance(other, flipper.IntegerType):
			return int(self) // other
		else:
			try:
				accuracy_needed = int(self.height + self.log_degree + other.height + other.log_degree) + 5
				return int(self.algebraic_approximation(accuracy_needed) / other.algebraic_approximation(accuracy_needed))
			except AttributeError:
				return NotImplemented
	def __mod__(self, other):
		if isinstance(other, flipper.IntegerType):
			return self.number_field.element([coeff % other for coeff in self])
		else:
			return NotImplemented
	
	def polynomial(self):
		''' Return a polynomial that this algebraic number is a root of.
		
		Note that this is NOT guaranteed to return the minimal polynomial
		of self. However it will whenever:
			deg(self) == def(self.number_field). '''
		
		# We get such a polynomial from the characteristic polynomial of the matrix
		# describing the action of self on the basis 1, \lmbda, \lambda^2, ...
		
		return flipper.kernel.dot(self, self.number_field.companion_matrices).characteristic_polynomial()
	
	def minimal_polynomial(self):
		''' Return the minimal polynomial of this algebraic number. '''
		
		# Get a really good approximation. Exactly how good we need is "...subtle, and
		# depends in part on what one knows about the problem" see Cohen Page 100.
		t = int(self.degree + self.height) + 1
		while True:
			I = self.interval_approximation(t)
			f = I.polynomial(self.degree, scale=10**t)
			
			if f(self) == 0:
				break
			else:
				t = 2 * t
		return f
	
	def compare(self, other, operator):
		''' Return the result of operator on sufficiently good algebraic approximations of self and other. '''
		
		if isinstance(other, flipper.IntegerType):
			accuracy_needed = int(self.height + self.log_degree + flipper.kernel.height_int(other) + 0) + 5
			return operator(self.algebraic_approximation(accuracy_needed), other)
		else:
			try:
				accuracy_needed = int(self.height + self.log_degree + other.height + other.log_degree) + 5
				return operator(self.algebraic_approximation(accuracy_needed), other.algebraic_approximation(accuracy_needed))
			except AttributeError:
				return NotImplemented
	
	def __lt__(self, other):
		return self.compare(other, lambda x, y: x < y)
	def __eq__(self, other):
		return self.compare(other, lambda x, y: x == y)
	def __gt__(self, other):
		return self.compare(other, lambda x, y: x > y)
	def __le__(self, other):
		return not (self > other)
	def __ne__(self, other):
		return not (self == other)
	def __ge__(self, other):
		return not (self < other)
	
	def interval_approximation(self, accuracy=0):
		''' Return an interval containing this algebraic number, correct to at least the requested accuracy. '''
		
		min_accuracy = 0
		target_accuracy = max(accuracy, min_accuracy)
		
		if self._interval is None or self._interval.accuracy < target_accuracy:
			# We need to request a little more accuracy for the intervals of \lmbda**i because of the sum.
			# We need at least: max(flipper.kernel.height_int(coefficient) for coefficient in self) + self.degree
			# more, but that involves a lot of logs so we'll use:
			request_accuracy = target_accuracy + self.height
			intervals = self.number_field.lmbda_approximations(request_accuracy)
			self._interval = flipper.kernel.dot(self, intervals)
			assert(self._interval.accuracy >= target_accuracy)
		
		return self._interval
	
	def algebraic_approximation(self, accuracy=None):
		''' Return an AlgebraicApproximation of this element which is correct to at least the requested accuracy.
		
		The accuracy returned is at least the minimum accuracy needed to determine a unique algebraic number. '''
		
		# Let:
		# Suppose that L is generator of N and this algebraic number is:
		#   a_0 + a_1 L + ... + a_D L^D.
		#   = a_0 + (a_1 + ( ... (a_{D-1} + a_D L) ... ) L ) L.
		# Then by induction on D:
		#   h(self) <= sum(h(a_i)) + d (h(L) + lg(2)) and
		#   deg(self) <= D.
		if accuracy is None: accuracy = 0
		min_accuracy = self.height + self.log_degree
		target_accuracy = max(accuracy, min_accuracy)
		
		request_accuracy = target_accuracy
		
		return flipper.kernel.AlgebraicApproximation(self.interval_approximation(request_accuracy), self.log_degree, self.height)





import flipper

import importlib

def main(verbose=False):
	total_time = 0
	for profile_name in dir(flipper.profile):
		if not profile_name.startswith('_'):
			profile = importlib.import_module('flipper.profile.%s' % profile_name)
			print('Running %s profile...' % profile_name)
			time = profile.main(verbose=verbose)
			total_time += time
			print('\tRan in %0.3fs' % time)
	
	print('Finished profiling.')
	print('\tTotal time: %0.3fs' % total_time)

if __name__ == '__main__':
	main()





''' A module for representing and manipulating maps between Triangulations.

Provides one class: Encoding. '''

import flipper

try:
	from Queue import Queue
except ImportError:
	from queue import Queue
from itertools import product

NT_TYPE_PERIODIC = 'Periodic'
NT_TYPE_REDUCIBLE = 'Reducible'  # Strictly this  means "reducible and not periodic".
NT_TYPE_PSEUDO_ANOSOV = 'Pseudo-Anosov'

class Encoding(object):
	''' This represents a map between two Triagulations.
	
	If it maps to and from the same triangulation then it represents
	a mapping class. This can be checked using self.is_mapping_class().
	
	The map is given by a sequence of EdgeFlips, LinearTransformations
	and Isometries which act from right to left.
	
	>>> import flipper
	>>> S = flipper.load('S_1_1')
	>>> aB = S.mapping_class('aB')
	>>> bA = S.mapping_class('bA')
	>>> ab = S.mapping_class('ab')
	>>> i = S.mapping_class('')
	>>> a = S.mapping_class('a')
	>>> a
	a
	>>> x = S.triangulation.encode([1])
	>>> x
	[Flip 1]
	'''
	def __init__(self, sequence, _cache=None):
		assert(isinstance(sequence, (list, tuple)))
		assert(len(sequence) > 0)
		assert(all(isinstance(item, flipper.kernel.Move) for item in sequence))
		# We used to also test:
		#  assert(all(x.source_triangulation == y.target_triangulation for x, y in zip(sequence, sequence[1:])))
		# However this makes composing Encodings a quadratic time algorithm!
		
		self.sequence = sequence
		
		self.source_triangulation = self.sequence[-1].source_triangulation
		self.target_triangulation = self.sequence[0].target_triangulation
		self.zeta = self.source_triangulation.zeta
		
		self._cache = {'name': ''} if _cache is None else _cache  # For caching hard to compute results.
	
	def without_cache(self):
		return Encoding(self.sequence, _cache={'name': self._cache['name']})
	
	def is_mapping_class(self):
		''' Return if this encoding is a mapping class.
		
		That is, if it maps to the triangulation it came from.
		
		>>> aB.is_mapping_class(), bA.is_mapping_class()
		(True, True)
		>>> x.is_mapping_class()
		False
		'''
		
		return self.source_triangulation == self.target_triangulation
	
	def __repr__(self):
		return str(self)
	def __str__(self):
		return self._cache['name'] if 'name' in self._cache else str(self.sequence)  # A backup name.
	def __iter__(self):
		return iter(self.sequence)
	def __len__(self):
		return len(self.sequence)
	def package(self):
		''' Return a small amount of info that self.source_triangulation can use to reconstruct this triangulation. '''
		return [item.package() for item in self]
	def __reduce__(self):
		return (create_encoding, (self.source_triangulation, self.package(), self._cache))
	def flip_length(self):
		''' Return the number of flips needed to realise this sequence. '''
		
		return sum(item.flip_length for item in self)
	def __getitem__(self, value):
		if isinstance(value, slice):
			# It turns out that handling all slices correctly is really hard.
			# We need to be very careful with "empty" slices. As Encodings require
			# non-empty sequences, we have to return just the id_encoding. This
			# ensures the Encoding that we return satisfies:
			#   self == self[:i] * self[i:j] * self[j:]
			# even when i == j.
			
			start = 0 if value.start is None else value.start if value.start >= 0 else len(self) + value.start
			stop = len(self) if value.stop is None else value.stop if value.stop >= 0 else len(self) + value.stop
			if start == stop:
				if 0 <= start < len(self):
					return self.sequence[start].target_triangulation.id_encoding()
				else:
					raise IndexError('list index out of range')
			return Encoding(self.sequence[value])
		elif isinstance(value, flipper.IntegerType):
			return self.sequence[value]
		else:
			return NotImplemented
	
	def __eq__(self, other):
		if isinstance(other, Encoding):
			if self.source_triangulation != other.source_triangulation or \
				self.target_triangulation != other.target_triangulation:
				raise ValueError('Cannot compare Encodings between different triangulations.')
			
			return all(self(curve) == other(curve) for curve in self.source_triangulation.key_curves())
		else:
			return NotImplemented
	def __ne__(self, other):
		return not (self == other)
	def is_homologous_to(self, other):
		''' Return if this encoding is homologous to other.
		
		Two maps are homologous if and only if they induce the same map
		from H_1(source_triangulation) to H_1(target_triangulation).
		
		>>> aB.is_homologous_to(aB), aB.is_homologous_to(bA), bA.is_homologous_to(ab), aB.is_homologous_to(i)
		(True, False, False, False)
		'''
		
		if isinstance(other, Encoding):
			if self.source_triangulation != other.source_triangulation or \
				self.target_triangulation != other.target_triangulation:
				raise ValueError('Cannot compare Encodings between different triangulations.')
			
			return all(self(curve).is_homologous_to(other(curve)) for curve in self.source_triangulation.key_curves())
		else:
			return NotImplemented
	
	def __call__(self, other):
		if isinstance(other, flipper.kernel.Lamination):
			if self.source_triangulation != other.triangulation:
				raise ValueError('Cannot apply an Encoding to a Lamination on a triangulation other than source_triangulation.')
			
			geometric = other.geometric
			algebraic = other.algebraic
			
			for item in reversed(self.sequence):
				geometric = item.apply_geometric(geometric)
				algebraic = item.apply_algebraic(algebraic)
			
			return self.target_triangulation.lamination(geometric, algebraic, remove_peripheral=False)
		else:
			return NotImplemented
	def __mul__(self, other):
		if isinstance(other, Encoding):
			if self.source_triangulation != other.target_triangulation:
				raise ValueError('Cannot compose Encodings over different triangulations.')
			
			return Encoding(self.sequence + other.sequence, _cache=dict() if 'name' not in self._cache or 'name' not in other._cache else {'name': self._cache['name'] + '.' + other._cache['name']})
		else:
			return NotImplemented
	def __pow__(self, k):
		assert(self.is_mapping_class())
		
		if k == 0:
			return self.source_triangulation.id_encoding()
		elif k > 0:
			return Encoding(self.sequence * k, _cache=dict() if 'name' not in self._cache else {'name': '(%s)^%d' % (self._cache['name'], k)})
		else:
			return self.inverse()**abs(k)
	
	def inverse(self):
		''' Return the inverse of this encoding.
		
		>>> aB.inverse() == bA, ab == ab.inverse(), i == i.inverse()
		(True, False, True)
		'''
		
		return Encoding([item.inverse() for item in reversed(self.sequence)], _cache=dict() if 'name' not in self._cache else {'name': '(%s)^-1' % self._cache['name']})
	
	def closing_isometries(self):
		''' Return all the possible isometries from self.target_triangulation to self.source_triangulation.
		
		These are the maps that can be used to close this into a mapping class. '''
		
		return self.target_triangulation.isometries_to(self.source_triangulation)
	
	def order(self):
		''' Return the order of this mapping class.
		
		If this has infinite order then return 0.
		
		This encoding must be a mapping class.
		
		>>> aB.order(), a.order()
		(0, 0)
		>>> i.order(), ab.order()
		(1, 6)
		'''
		
		assert(self.is_mapping_class())
		
		# We could do:
		# for i in range(1, self.source_triangulation.max_order + 1):
		#	if self**i == self.source_triangulation.id_encoding():
		#		return i
		# But this is quadratic in the order so instead we do:
		curves = self.source_triangulation.key_curves()
		possible_orders = set(range(1, self.source_triangulation.max_order+1))
		for curve in curves:
			curve_image = curve
			for i in range(1, max(possible_orders)+1):
				curve_image = self(curve_image)
				if curve_image != curve:
					possible_orders.discard(i)
					if not possible_orders: return 0  # No finite orders remain so we are infinite order.
		
		return min(possible_orders)
	
	def is_identity(self):
		''' Return if this encoding is the identity map.
		
		>>> i.is_identity()
		True
		>>> aB.is_identity()
		False
		'''
		
		return self.is_mapping_class() and all(self(curve) == curve for curve in self.source_triangulation.key_curves())
	
	def is_periodic(self):
		''' Return if this encoding has finite order.
		
		This encoding must be a mapping class.
		
		>>> aB.is_periodic(), a.is_periodic()
		(False, False)
		>>> i.is_periodic(), ab.is_periodic()
		(True, True)
		'''
		
		return self.order() > 0
	
	def is_reducible(self):
		''' Return if this encoding is reducible and NOT periodic.
		
		This encoding must be a mapping class.
		
		>>> aB.is_reducible(), a.is_reducible()
		(False, True)
		>>> i.is_reducible(), ab.is_reducible()
		(False, False)
		'''
		
		return self.nielsen_thurston_type() == NT_TYPE_REDUCIBLE
	
	def is_pseudo_anosov(self):
		''' Return if this encoding is pseudo-Anosov.
		
		This encoding must be a mapping class.
		
		>>> aB.is_pseudo_anosov(), a.is_pseudo_anosov()
		(True, False)
		>>> i.is_pseudo_anosov(), ab.is_pseudo_anosov()
		(False, False)
		'''
		
		return self.nielsen_thurston_type() == NT_TYPE_PSEUDO_ANOSOV
	
	def applied_geometric(self, lamination):
		''' Return the action and condition matrices describing the PL map
		applied to the geometric coordinates of the given lamination. '''
		
		assert(isinstance(lamination, flipper.kernel.Lamination))
		
		As = flipper.kernel.id_matrix(self.zeta)
		Cs = flipper.kernel.zero_matrix(self.zeta, 1)
		for item in reversed(self.sequence):  # We can't just reverse self as reversed requires a list not an iterator.
			# This now uses the improved sparse matrix representation.
			As, C = item.applied_geometric(lamination, As)
			Cs = Cs.join(C)
			lamination = item(lamination)
		
		return As, Cs
	
	def pl_action(self):
		''' Yield each of the action, condition matrix pairs describing the action of this Encoding
		on ML. '''
		
		for sequence in product(*[range(len(item)) for item in self]):
			As = flipper.kernel.id_matrix(self.zeta)
			Cs = flipper.kernel.zero_matrix(self.zeta, 1)
			for item, index in reversed(list(zip(self, sequence))):
				# This now uses the improved sparse matrix representation.
				As, C = item.pl_action(index, action=As)
				Cs = Cs.join(C)
		
			yield (As, Cs)
		
		return
	
	def pml_fixedpoint_uncached(self, starting_curve=None):
		''' Return a rescaling constant and projectively invariant lamination.
		
		Assumes that the mapping class is pseudo-Anosov.
		
		To find this we start with a curve on the surface and repeatedly apply
		the map until it appear to be projectively similar to a previous iteration.
		Finally it uses:
			flipper.kernel.symboliccomputation.perron_frobenius_eigen()
		to find the nearby projective fixed point. If this fails after some number
		of iterations then we fall back to computing the action inside of every
		cell of the PL action. Generically this is exponentially slow.
		
		Note: In most pseudo-Anosov cases < 15 iterations are needed, if it fails
		to converge after many iterations then it is extremely likely that the
		mapping class was not pseudo-Anosov. Also note that the number of iterations
		done before switching technique is (currently) only dependent on the topology
		of the underlying surface.
		
		This encoding must be a mapping class. '''
		
		# Suppose that f_1, ..., f_m, g_1, ..., g_n, t_1, ..., t_k, p is the Thurston decomposition of self.
		# That is: f_i are pA on subsurfaces, g_i are periodic on subsurfaces, t_i are Dehn twist along the curve of
		# the canonical curve system and p is a permutation of the subsurfaces.
		# Additionally, let S_i be the subsurface corresponding to f_i, P_i correspond to g_i and A_i correspond to t_i.
		# Finally, let x_0 be a curve on the surface and define x_i := self(x_{i-1}).
		#
		# The algorithm covers 3 cases:  (Note we reorder the subsurfaces for ease of notation.)
		#  1) x_0 meets at S_1, ..., S_m',
		#  2) x_0 meets no S_i but meets A_1, ..., A_k', and
		#  3) x_0 meets no S_i or A_i, that is x_0 is contained in a P_1.
		#
		# In the first case, x_i will converge exponentially to the stable laminations of f_1, ..., f_m'.
		# Here the convergence is so fast we need only a few iterations.
		#
		# In the second case x_i will converge linearly to c, the cores of A_1, ..., A_k'. To speed this up
		# we note that x_i = i*c + O(1), so rescale x_i by 1/i, round and check if this is c.
		#
		# Finally, the third case happens if and only if x_i is periodic. In this case self must be
		# periodic or reducible. We test for periodicity at the beginning hence if we ever find a curve
		# fixed by a power of self then we must reducible.
		
		# Possible future improvements:
		#  - Store the checked cells in a hash map to prevent rerunning the expensive symboliccomputation.directed_eigenvector.
		#  - Automatically update to a finer RESOLUTION whenever you fail to get a lamination from symboliccomputation.directed_eigenvector.
		
		assert(self.is_mapping_class())
		
		# We will use a hash to significantly speed up the algorithm.
		resolution = 200
		def curve_hash(curve, resolution):
			''' A simple hash mapping cuves to a coarse lattice in PML. '''
			# Hmmm, this can suffer from // always rounding down.
			w = curve.weight()
			return (resolution,) + tuple([entry * resolution // w for entry in curve])
		
		# We start with a fast test for periodicity.
		# This isn't needed but it means that if we ever discover that
		# self is not pA then it must be reducible.
		if self.is_periodic():
			raise flipper.AssumptionError('Mapping class is periodic.')
		
		triangulation = self.source_triangulation
		max_order = triangulation.max_order
		if starting_curve is None: starting_curve = triangulation.key_curves()[0]
		curves = [starting_curve]
		seen = {curve_hash(curves[0], resolution): [0]}
		tested = set()
		# Experimentally this is a good number to do.
		# Should it be max_order**2 or even depend on len(self)?
		for i in range(max(10 * max_order, 100)):
			new_curve = self(curves[-1])
			curves.append(new_curve)
			
			hsh = curve_hash(new_curve, resolution)
			# print(i, new_curve, hsh)
			if hsh in seen:
				for j in reversed(seen[hsh]):  # Better to work backwards as the later ones are likely to be longer and so projectively closer.
					# Check if we have seen this curve before.
					if new_curve == curves[j]:  # self**(i-j)(new_curve) == new_curve, so self is reducible.
						raise flipper.AssumptionError('Mapping class is reducible.')
					# Average the last few curves in case they have 'spiralled' around the fixedpoint.
					average_curve = sum(curves[j:])
					action_matrix, condition_matrix = self.applied_geometric(average_curve)
					# We'll store condition_matrix in a set to ensure that we never test the same cell twice.
					if condition_matrix not in tested:
						tested.add(condition_matrix)
						try:
							eigenvalue, eigenvector = flipper.kernel.symboliccomputation.directed_eigenvector(
								action_matrix, condition_matrix)
						except flipper.ComputationError:
							pass  # Could not find an eigenvector in the cone.
						except flipper.AssumptionError:
							raise flipper.AssumptionError('Mapping class is reducible.')
						else:
							invariant_lamination = triangulation.lamination(eigenvector)
							if not invariant_lamination.is_empty():  # But it might have been entirely peripheral.
								return eigenvalue, invariant_lamination
				
				seen[hsh].append(i+1)
			else:
				seen[hsh] = [i+1]
			
			# We now have an extra test to handle the case when self is reducible and curve lies only in periodic parts.
			if len(seen[hsh]) > 4 or i % max_order == 0:  # This is still slow (quadratic in max_order) so don't do it often.
				for j in reversed(range(max(len(curves) - max_order, 0), len(curves)-1)):
					# A good guess for the reducing curve is the (additive) growth rate.
					vector = [x - y for x, y in zip(new_curve, curves[j])]
					new_small_curve = small_curve = triangulation.lamination(vector, algebraic=[0] * self.zeta)
					if not small_curve.is_empty():
						for k in range(1, max_order+1):
							new_small_curve = self(new_small_curve)
							if new_small_curve == small_curve:
								if k == 1:
									# We could raise an AssumptionError in this case too as this also shows that self is reducible.
									return 1, small_curve
								else:
									raise flipper.AssumptionError('Mapping class is reducible.')
			
			if len(seen[hsh]) > 6:
				# Recompute seen to a higher resolution.
				# This reduces the chances that we will get false positives that need
				# to have an expensive directed_eigenvector calculation done on them.
				resolution = resolution * 10  # Crank up exponentially.
		
		# If all else fails then just check every cell of the PL action. There can be exponentially many cells
		# (and generically there are) so this process is extremely slow. However it is guaranteed to terminate
		# with an invariant lamination or prove that the mapping class is reducible.
		for action_matrix, condition_matrix in self.pl_action():
			if condition_matrix not in tested:  # We may as well skip cells that we've already tested.
				try:
					eigenvalue, eigenvector = flipper.kernel.symboliccomputation.directed_eigenvector(
						action_matrix, condition_matrix)
				except flipper.ComputationError:
					pass  # Could not find an eigenvector in the cone.
				except flipper.AssumptionError:
					raise flipper.AssumptionError('Mapping class is reducible.')
				else:
					invariant_lamination = triangulation.lamination(eigenvector)
					if not invariant_lamination.is_empty():  # But it might have been entirely peripheral.
						return eigenvalue, invariant_lamination
		
		# If none of the cells contained a projective fixed-point with eigenvalue > 1 then this mapping class
		# must be reducible.
		raise flipper.AssumptionError('Mapping class is reducible.')
	
	def pml_fixedpoint(self):
		''' A version of self.invariant_lamination_uncached with caching. '''
		
		if 'invariant_lamination' not in self._cache:
			try:
				self._cache['invariant_lamination'] = self.pml_fixedpoint_uncached()
			except flipper.AssumptionError as error:
				self._cache['invariant_lamination'] = error
		
		if isinstance(self._cache['invariant_lamination'], Exception):
			raise self._cache['invariant_lamination']
		else:
			return self._cache['invariant_lamination']
	
	def invariant_lamination(self):
		''' Return a projectively invariant lamination of this mapping class.
		
		This encoding must be a mapping class.
		
		Assumes that this encoding is pseudo-Anosov. '''
		
		_, lamination = self.pml_fixedpoint()
		return lamination
	
	def dilatation(self):
		''' Return the dilatation of this mapping class.
		
		This encoding must be a mapping class. '''
		
		if self.nielsen_thurston_type() != NT_TYPE_PSEUDO_ANOSOV:
			return flipper.kernel.NumberField().one  # This is equivalent to ZZ[1].one.
		else:
			lmbda, _ = self.pml_fixedpoint()
			return lmbda
	
	def splitting_sequences(self, take_roots=False):
		''' Return a list of splitting sequences associated to this mapping class.
		
		Assumes (and checks) that the mapping class is pseudo-Anosov.
		
		This encoding must be a mapping class. '''
		
		if self.is_periodic():  # Actually this test is redundant but it is faster to test it now.
			raise flipper.AssumptionError('Mapping class is not pseudo-Anosov.')
		
		dilatation, lamination = self.pml_fixedpoint()
		try:
			splittings = lamination.splitting_sequences(dilatation=None if take_roots else dilatation)
		except flipper.AssumptionError:  # Lamination is not filling.
			raise flipper.AssumptionError('Mapping class is not pseudo-Anosov.')
		
		return splittings
	
	def splitting_sequence(self):
		''' Return the splitting sequence associated to this mapping class.
		
		Assumes (and checks) that the mapping class is pseudo-Anosov.
		
		This encoding must be a mapping class. '''
		
		# We get a list of all possible splitting sequences from
		# self.splitting_sequences(). From there we use the fact that each
		# of these differ by a periodic mapping class, which cannot be in the
		# Torelli subgroup and so acts non-trivially on H_1(S).
		# Thus we look for the map whose action on H_1(S) is conjugate to self
		# via splitting.preperiodic.
		
		# To do this we take sufficiently many curves (the key_curves() of the
		# underlying triangulation) and look for the splitting sequence in which
		# they are mapped to homologous curves by:
		#	preperiodic . self and periodic . preperiodic.
		# Note that we no longer use self.inverse() as periodic now goes in the
		# same direction as self.
		
		homology_splittings = [splitting for splitting in self.splitting_sequences()
			if (splitting.preperiodic * self).is_homologous_to(splitting.mapping_class * splitting.preperiodic)]
		
		if len(homology_splittings) == 0:
			raise flipper.FatalError('Mapping class is not homologous to any splitting sequence.')
		elif len(homology_splittings) == 1:
			return homology_splittings[0]
		else:
			raise flipper.FatalError('Mapping class is homologous to multiple splitting sequences.')
	
	def canonical(self):
		''' Return the canonical form of this mapping class. '''
		
		return self.splitting_sequence().mapping_class
	
	def nielsen_thurston_type(self):
		''' Return the Nielsen--Thurston type of this encoding.
		
		This encoding must be a mapping class.
		
		>>> ab.nielsen_thurston_type(), a.nielsen_thurston_type(), aB.nielsen_thurston_type()
		('Periodic', 'Reducible', 'Pseudo-Anosov')
		'''
		
		if self.is_periodic():
			return NT_TYPE_PERIODIC
		
		try:
			# We could do any of self.splitting_sequence(), self.canonical(), ...
			# but this involves the least calculation and so is fastest.
			self.splitting_sequences()
		except flipper.AssumptionError:
			return NT_TYPE_REDUCIBLE
		
		return NT_TYPE_PSEUDO_ANOSOV
	
	def is_abelian(self):
		''' Return if this mapping class corresponds to an Abelian differential.
		
		This is an Abelian differential (rather than a quadratic differential) if and
		only if its stable lamination is orientable.
		
		Assumes (and checks) that the mapping class is pseudo-Anosov.
		
		This encoding must be a mapping class.
		
		>>> aB.is_abelian()
		True
		>>> ab.is_abelian()  # doctest: +ELLIPSIS
		Traceback (most recent call last):
		    ...
		AssumptionError: ...
		>>> a.is_abelian()  # doctest: +ELLIPSIS
		Traceback (most recent call last):
		    ...
		AssumptionError: ...
		'''
		
		return self.splitting_sequence().lamination.is_orientable()
	
	def is_primitive(self):
		''' Return if this mapping class is primitive.
		
		This encoding must be a mapping class.
		
		Assumes (and checks) that this mapping class is pseudo-Anosov. '''
		
		return self.splitting_sequences(take_roots=True).open_periodic.flip_length() == self.canonical().flip_length()
	
	def is_conjugate_to(self, other):
		''' Return if this mapping class is conjugate to other.
		
		It would also be straightforward to check if self^i ~~ other^j
		for some i, j.
		
		Both encodings must be mapping classes.
		
		Currently assumes that at least one mapping class is pseudo-Anosov. '''
		
		assert(isinstance(other, Encoding))
		
		# Nielsen-Thurston type is a conjugacy invariant.
		if self.nielsen_thurston_type() != other.nielsen_thurston_type():
			return False
		
		if self.nielsen_thurston_type() == NT_TYPE_PERIODIC:
			if self.order() != other.order():
				return False
			
			# We could also use action on H_1(S) as a conjugacy invaraiant.
			
			raise flipper.AssumptionError('Mapping class is periodic.')
		elif self.nielsen_thurston_type() == NT_TYPE_REDUCIBLE:
			# There's more to do here.
			
			raise flipper.AssumptionError('Mapping class is reducible.')
		elif self.nielsen_thurston_type() == NT_TYPE_PSEUDO_ANOSOV:
			# Two pseudo-Anosov mapping classes are conjugate if and only if
			# there canonical forms are cyclically conjugate via an isometry.
			f = self.canonical()
			g = other.canonical()
			# We should start by quickly checking some invariants.
			# For example they should have the same dilatation.
			if self.dilatation() != other.dilatation():
				return False
			
			for i in range(len(f)):
				# Conjugate around.
				f_cycled = f[i:] * f[:i]
				# g_cycled = g[i:] * g[:i]  # Could cycle g instead.
				for isom in f_cycled.source_triangulation.isometries_to(g.source_triangulation):
					if isom.encode() * f_cycled == g * isom.encode():
						return True
			
			return False
	
	def stratum(self):
		''' Return a dictionary mapping each singularity to its stratum order.
		
		This is the number of bipods incident to the vertex.
		
		Assumes (and checks) that this mapping class is pseudo-Anosov.
		
		This encoding must be a mapping class. '''
		
		# This can fail with an flipper.AssumptionError.
		return self.splitting_sequence().lamination.stratum()
	
	def hitting_matrix(self):
		''' Return the hitting matrix of the underlying train track. '''
		
		# This can fail with an flipper.AssumptionError.
		h = self.canonical()
		
		M = flipper.kernel.id_matrix(h.zeta)
		lamination = h.invariant_lamination()
		# Lamination defines a train track with a bipod in each triangle. We
		# follow the sequence of folds (and isometries) which this train track
		# undergoes and track how the edges are mapped using M.
		for item in reversed(h.sequence):
			if isinstance(item, flipper.kernel.EdgeFlip):
				triangulation = lamination.triangulation
				a, b, c, d = triangulation.square_about_edge(item.edge_label)
				
				# Work out which way the train track is pointing.
				if lamination.is_bipod(triangulation.corner_of_edge(a.label)):
					assert(lamination.is_bipod(triangulation.corner_of_edge(c.label)))
					M = M.elementary(item.edge_index, b.index)
					M = M.elementary(item.edge_index, d.index)
				elif lamination.is_bipod(triangulation.corner_of_edge(b.label)):
					assert(lamination.is_bipod(triangulation.corner_of_edge(d.label)))
					M = M.elementary(item.edge_index, a.index)
					M = M.elementary(item.edge_index, c.index)
				else:
					raise flipper.FatalError('Incompatible bipod.')
			elif isinstance(item, flipper.kernel.Isometry):
				M = flipper.kernel.Permutation([item.index_map[i] for i in range(h.zeta)]).matrix() * M
			else:
				raise flipper.FatalError('Unknown item in canonical sequence.')
			
			# Move the lamination onto the next triangulation.
			lamination = item(lamination)
		
		return M
	
	def bundle(self, canonical=True, _safety=True):
		''' Return the bundle associated to this mapping class.
		
		This method can be run in two different modes:
		If canonical=True:
			Then the bundle returned is triangulated by a veering, layered
			triangulation and has at most 6g+5n-6 additional loops drilled
			from it, as described by Agol. These additional cusps are marked
			as fake cusps and can be dealt with by filling along their
			fibre slope.
			
			Assumes (and checks) that this mapping class is pseudo-Anosov.
		If canonical=False:
			Then the bundle returned is triangulated by a layered triangulation
			obtained by stacking flat tetrahedra, one for each edge flip in
			self.
			
			Assumes (and checks) that the resulting triangulation is an ideal
			triangulation of a manifold and that the fibre surface immerses
			into the two skeleton. If _safety=True then this should always happen.
		
		This encoding must be a mapping class. '''
		
		assert(self.is_mapping_class())
		triangulation = self.source_triangulation
		
		if canonical:
			# This can fail with an flipper.AssumptionError if self is not pseudo-Anosov.
			return self.canonical().bundle(canonical=False, _safety=False)
		
		if _safety:
			# We should add enough flips to ensure the triangulation is a manifold.
			# Flipping and then unflipping every edge is certainly enough.
			# However, we still have to be careful as there may be non-flippable edges.
			
			# Start by adding a flip and unflip each flippable edge.
			safe_encoding = self
			for i in triangulation.flippable_edges():
				extra = triangulation.encode_flip(i)
				safe_encoding = extra.inverse() * extra * safe_encoding
			# Then add a flip and unflip for each non-flippable edge.
			for i in triangulation.indices:
				if not triangulation.is_flippable(i):
					# To do this we must first flip the boundary edge.
					boundary_edge = triangulation.nonflippable_boundary(i)
					# This edge is always flippable and, after flipping it, i is too.
					extra = triangulation.encode([i, boundary_edge])
					safe_encoding = extra.inverse() * extra * safe_encoding
			
			return safe_encoding.bundle(canonical=False, _safety=False)
		
		VEERING_LEFT, VEERING_RIGHT = flipper.kernel.triangulation3.VEERING_LEFT, flipper.kernel.triangulation3.VEERING_RIGHT
		id_perm3 = flipper.kernel.Permutation((0, 1, 2))
		
		
		lower_triangulation, upper_triangulation = triangulation, triangulation
		
		triangulation3 = flipper.kernel.Triangulation3(self.flip_length())
		# These are maps taking triangles of lower (respectively upper) triangulation to either:
		#  - A pair (triangle, permutation) where triangle is in upper (resp. lower) triangulation, or
		#  - A pair (tetrahedron, permutation) of triangulation3.
		# We start with no tetrahedra, so these maps are just the identity map between the two triangulations.
		lower_map = dict((triangleA, (triangleB, id_perm3)) for triangleA, triangleB in zip(lower_triangulation, upper_triangulation))
		upper_map = dict((triangleB, (triangleA, id_perm3)) for triangleA, triangleB in zip(lower_triangulation, upper_triangulation))
		
		# We also use these two functions to quickly tell what a triangle maps to.
		maps_to_triangle = lambda X: isinstance(X[0], flipper.kernel.Triangle)
		maps_to_tetrahedron = lambda X: not maps_to_triangle(X)
		
		tetra_count = 0
		for item in reversed(self.sequence):
			assert(item.source_triangulation == upper_triangulation)
			
			try:
				tetra_count, upper_triangulation, upper_map, lower_map = \
					item.extend_bundle(triangulation3, tetra_count, upper_triangulation, lower_triangulation, upper_map, lower_map)
			except AttributeError:
				# We have no way to handle any other type that appears.
				# Currently this means there was a LinearTransform and so this is not a mapping class.
				raise flipper.FatalError('Unknown move %s encountered while building bundle.' % item)
		
		# We're now back to the starting triangulation.
		assert(lower_triangulation == upper_triangulation)
		
		# This is a map which send each triangle of upper_triangulation via isometry to a pair:
		#	(triangle, permutation)
		# where triangle in lower_triangulation and maps_to_tetrahedron(lower_map[triangle]).
		full_forwards = dict()
		for source_triangle in upper_triangulation:
			target_triangle, perm = source_triangle, id_perm3
			
			c = 0
			while maps_to_triangle(lower_map[target_triangle]):
				target_triangle, new_perm = lower_map[target_triangle]
				perm = new_perm * perm
				
				c += 1
				assert(c <= 3 * upper_triangulation.zeta)
			full_forwards[source_triangle] = (target_triangle, perm)
		
		# Now close the bundle up.
		for source_triangle in upper_triangulation:
			if maps_to_tetrahedron(upper_map[source_triangle]):
				A, perm_A = upper_map[source_triangle]
				target_triangle, perm = full_forwards[source_triangle]
				B, perm_B = lower_map[target_triangle]
				A.glue(perm_A(3), B, perm_B * perm.embed(4) * perm_A.inverse())
		
		# There are now no unglued faces.
		assert(triangulation3.is_closed())
		
		# Install longitudes and meridians. This also calls Triangulation3.assign_cusp_indices().
		triangulation3.install_peripheral_curves()
		
		# Construct an immersion of the fibre surface into the closed bundle.
		fibre_immersion = dict()
		for source_triangle in lower_triangulation:
			if maps_to_triangle(lower_map[source_triangle]):
				upper_triangle, upper_perm = lower_map[source_triangle]
				target_triangle, perm = full_forwards[upper_triangle]
				B, perm_B = lower_map[target_triangle]
				fibre_immersion[source_triangle] = (B, perm_B * (perm * upper_perm).embed(4))
			else:
				B, perm_B = lower_map[source_triangle]
				fibre_immersion[source_triangle] = lower_map[source_triangle]
		
		return flipper.kernel.Bundle(lower_triangulation, triangulation3, fibre_immersion)
	
	def flat_structure(self):
		''' Return the flat structure associated to self.canonical().
		
		Assumes that this mapping class is pseudo-Anosov.
		
		This encoding must be a mapping class. '''
		
		assert(self.is_mapping_class())
		
		stable_lamination = self.splitting_sequence().lamination
		unstable_lamination = self.splitting_sequence().mapping_class.inverse().invariant_lamination()
		periodic_triangulation = self.splitting_sequence().triangulation
		
		queue = Queue()  # The edges whose vectors are decided but whose triangles have not been inspected.
		edge_vectors = dict()  # A dictionary mapping edges to their corresponding vectors. 
		
		while len(edge_vectors) < self.zeta:
			# We fix an edge to start the procedure.
			e = [edge for edge in periodic_triangulation.edges if edge not in edge_vectors][0]
			
			# Fill in the basic information for this edge (and its inverse).
			edge_vectors[e] = flipper.kernel.Vector2(stable_lamination(e), unstable_lamination(e))
			edge_vectors[~e] = flipper.kernel.Vector2(-stable_lamination(e), -unstable_lamination(e))
			
			queue.put(e)
			queue.put(~e)
			
			while not queue.empty():
				edge_1 = queue.get()  # Get an unchecked edge.
				# Moving anticlockwise around the triangle from edge_1 brings you to edge_2 then edge_3
				#          /\
				#         /  \
				#        /    \
				#  e_3  v      ^  e_2
				#      /        \
				#     /          \
				#     ------>-----
				#         e_1
				corner = periodic_triangulation.corner_lookup[edge_1.label]
				edge_2, edge_3 = corner.edges[1], corner.edges[2]
				
				if edge_2 in edge_vectors and edge_3 not in edge_vectors:  # If two of the edges are known we can calculate the third using the formula: edge_1  + edge_2  + edge_3 == 0.
					edge_vectors[edge_3] = -(edge_vectors[edge_1] + edge_vectors[edge_2])
					edge_vectors[~edge_3] = -edge_vectors[edge_3]
					queue.put(~edge_3)
				elif edge_2 not in edge_vectors and edge_3 in edge_vectors:  # The other possibility if two edges are known.
					edge_vectors[edge_2] = -(edge_vectors[edge_1] + edge_vectors[edge_3])
					edge_vectors[~edge_2] = -edge_vectors[edge_2]
					queue.put(~edge_2)
				elif edge_2 not in edge_vectors and edge_3 not in edge_vectors:  # If edge_2 and edge_3 are both unknown:
					# Here we just choose positive signs in anticipation of changing them later if need be.
					s_1, u_1 = stable_lamination(edge_1), unstable_lamination(edge_1)
					s_2, u_2 = stable_lamination(edge_2), unstable_lamination(edge_2)
					s_3, u_3 = stable_lamination(edge_3), unstable_lamination(edge_3)
					
					edge_vectors[edge_2] = flipper.kernel.Vector2(
						-s_2 if (s_1 == s_2 + s_3 and edge_vectors[edge_1].x >= 0) or \
							(s_2 == s_1 + s_3 and edge_vectors[edge_1].x >= 0) or \
							(s_3 == s_1 + s_2 and edge_vectors[edge_1].x < 0) else s_2,
						-u_2 if (u_1 == u_2 + u_3 and edge_vectors[edge_1].y >= 0) or \
							(u_2 == u_1 + u_3 and edge_vectors[edge_1].y >= 0) or \
							(u_3 == u_1 + u_2 and edge_vectors[edge_1].y < 0) else u_2
						)
					
					edge_vectors[edge_3] = flipper.kernel.Vector2(
						-s_3 if (s_1 == s_2 + s_3 and edge_vectors[edge_1].x >= 0) or \
							(s_2 == s_1 + s_3 and edge_vectors[edge_1].x < 0) or \
							(s_3 == s_1 + s_2 and edge_vectors[edge_1].x >= 0) else s_3,
						-u_3 if (u_1 == u_2 + u_3 and edge_vectors[edge_1].y >= 0) or \
							(u_2 == u_1 + u_3 and edge_vectors[edge_1].y < 0) or \
							(u_3 == u_1 + u_2 and edge_vectors[edge_1].y >= 0) else u_3
						)
					
					# Put the inverses in too.
					edge_vectors[~edge_2] = -edge_vectors[edge_2]
					edge_vectors[~edge_3] = -edge_vectors[edge_3]
					queue.put(~edge_2)
					queue.put(~edge_3)
				# The final case is: edge_2 in edge_vectors and edge_3 in edge_vectors. However here there is nothing to do as all of the edges are known.
		
		return flipper.kernel.FlatStructure(periodic_triangulation, edge_vectors)

def create_encoding(source_triangulation, sequence, _cache=None):
	''' Return the encoding defined by sequence starting at source_triangulation.
	
	This is only really here to help with pickling. Users should use
	source_triangulation.encode(sequence) directly. '''
	
	assert(isinstance(source_triangulation, flipper.kernel.Triangulation))
	
	return source_triangulation.encode(sequence, _cache=_cache)

def doctest_globs():
	''' Return the globals needed to run doctest on this module. '''
	
	S = flipper.load('S_1_1')
	aB = S.mapping_class('aB')
	bA = S.mapping_class('bA')
	ab = S.mapping_class('ab')
	i = S.mapping_class('')
	a = S.mapping_class('a')
	x = S.triangulation.encode([1])
	
	return {'aB': aB, 'bA': bA, 'ab': ab, 'i': i, 'a': a, 'x': x}





from __future__ import print_function

import os
try:
	from setuptools import setup
except ImportError:
	print('Unable to import setuptools, using distutils instead.')
	from distutils.core import setup

this_directory = os.path.dirname(__file__)
source_directory = os.path.join(this_directory, 'source')
exec(open(os.path.join(source_directory, 'version.py')).read())  # Load in the variable __version__.

setup(
	name='flipper',
	version=__version__,
	description='For manipulating curves and measured laminations on surfaces and producing mapping tori.',
	author='Mark Bell',
	author_email='M.C.Bell@warwick.ac.uk',
	url='https://bitbucket.org/Mark_Bell/flipper',
	# Remember to update these if the directory structure changes.
	packages=[
		'flipper',
		'flipper.application',
		'flipper.doc',
		'flipper.example',
		'flipper.kernel',
		'flipper.kernel.interface',
		'flipper.test',
		'flipper.profile'
		],
	package_dir={'flipper': source_directory},
	package_data={
		'flipper': ['censuses/*.dat'],
		'flipper.doc': ['flipper.pdf'],
		'flipper.application': ['icon/*', 'frames/*'],
		},
	install_requires=['cypari']
	)





''' A simple starting point for flipper. '''

import flipper

def main():
	''' Describe how to start flipper. '''
	
	print('flipper %s' % flipper.__version__)
	print('Some basic commands:')
	print('  > python -m flipper.app      # To start the flipper GUI.')
	print('  > python -m flipper.doc      # To open the flipper documentation.')
	print('  > python -m flipper.example  # To see a list of included examples.')
	print('  > python -m flipper.test     # To test your installation of flipper.')
	print('  > python -m flipper.profile  # To see profile your installation of flipper.')
	print('Or start Python and import flipper directly.')

if __name__ == '__main__':
	main()





from __future__ import print_function

import flipper
try:
	import snappy
except ImportError:
	snappy = None

def main():
	while True:
		surface = raw_input('Choose surface (leave blank to quit): ')
		if surface == '': break
		try:
			S = flipper.load(surface)
		except KeyError:
			print('Not a valid surface name')
		else:
			while True:
				word = raw_input('Enter mapping class (leave blank to rechoose surface): ')
				if word == '': break
				try:
					M = snappy.twister.Surface(surface).bundle(word)
					mapping_class = S.mapping_class(word)
					match = M.is_isometric_to(snappy.Manifold(mapping_class.bundle().snappy_string()))
					print('Match: %s' % match)
				except KeyError:
					print('Not a valid mapping class.')
				except flipper.AssumptionError:
					print('Mapping class is not pseudo-Anosov.')

if __name__ == '__main__':
	import argparse
	parser = argparse.ArgumentParser(description='Compare bundles built with flipper and Twister')
	parser.add_argument('--show', action='store_true', default=False, help='show the source code of this example and exit')
	args = parser.parse_args()
	
	if args.show:
		print(open(__file__, 'r').read())
	else:
		if snappy is None:
			print('This example requires SnapPy.')
			print('Please install it and try again.')
		else:
			main()





''' A module for loading flipper databases. '''

import flipper

import os

DATABASE_DIRECTORY = os.path.join(os.path.dirname(__file__), 'censuses')

def find_databases():
	''' Return a list of all loadable databases. '''
	
	databases = []
	for path in os.listdir(DATABASE_DIRECTORY):
		_, tail = os.path.split(path)
		name, extension = os.path.splitext(tail)
		if extension == '.dat':
			databases.append(name)
	
	return databases

def census(census_name, find=None):
	''' Iterates through the requested database.
	
	If given, only results containing find are returned. '''
	
	assert(isinstance(census_name, flipper.StringType))
	assert(find is None or isinstance(find, flipper.StringType))
	
	if census_name in find_databases():
		census_name = os.path.join(DATABASE_DIRECTORY, census_name + '.dat')
	
	with open(census_name) as f:
		for line in f:
			if find is None or line.find(find) != -1:
				line = line.strip()
				if '#' in line:
					line = line[:line.find('#')]
				
				if line:
					yield line.strip().split('\t')





from __future__ import print_function
from time import time

import flipper

# This example shows that even for generic mapping classes Encoding.invariant_lamination() is fast.

def main(n=100):
	times = {}
	surface = 'S_3_1'
	S = flipper.load(surface)
	for index in range(n):
		word = S.random_word(20)  # , negative=False)
		print('%3d/%d: %s %s' % (index+1, n, surface, word.replace('.', '')), end='')
		mapping_class = S.mapping_class(word)
		t = time()
		try:
			mapping_class.invariant_lamination()
		except flipper.AssumptionError:
			print(', Claim: not pA', end='')
		times[word] = time() - t
		print(', Time: %0.3f' % times[word])
	print('Average time: %0.3f' % (sum(times.values()) / n))
	print('Slowest: %s, Time: %0.3f' % (max(times, key=lambda w: times[w]).replace('.', ''), max(times.values())))
	print('Total time: %0.3f' % sum(times.values()))

if __name__ == '__main__':
	import argparse
	parser = argparse.ArgumentParser(description='Demonstrate the effectiveness of flippers heuristics for finding invariant laminations.')
	parser.add_argument('--show', action='store_true', default=False, help='show the source code of this example and exit')
	args = parser.parse_args()
	
	if args.show:
		print(open(__file__, 'r').read())
	else:
		main()





try:
	import Tkinter as TK
except ImportError:  # Python 3.
	import tkinter as TK

class GetStringApp(object):
	def __init__(self, title, prompt, default='', validate=None, host_app_parent=None):
		if host_app_parent is None: host_app_parent = TK._default_root
		self.host_app_parent = host_app_parent
		self.validate = validate
		self.result = None
		self.parent = TK.Toplevel(self.host_app_parent)
		self.parent.withdraw()  # Hide self while we set up the geometry.
		self.parent.title(title)
		self.parent.protocol('WM_DELETE_WINDOW', self.cancel)  # To catch when users click the 'x' to close the window.
		
		self.text_label = TK.Label(self.parent, text=prompt, justify='left')
		
		self.button_frame = TK.Frame(self.parent)
		self.button_ok = TK.Button(self.button_frame, text='Ok', command=self.ok)
		self.button_cancel = TK.Button(self.button_frame, text='Cancel', command=self.cancel)
		self.button_ok.pack(pady=2, fill='x', expand=True)
		self.button_cancel.pack(pady=2, fill='x', expand=True)
		
		self.variable = TK.StringVar()
		self.text_entry = TK.Entry(self.parent, width=40, textvariable=self.variable)
		self.variable.set(default)
		self.text_entry.selection_range(0, TK.END)
		
		self.text_label.grid(row=0, column=0, padx=5, pady=5, sticky='nw')
		self.button_frame.grid(row=0, column=1, padx=5, pady=5, sticky='e')
		
		self.text_entry.grid(row=1, column=0, columnspan=2, padx=5, pady=5, sticky='we')
		
		self.parent.bind('<Return>', self.ok)
		self.parent.bind('<Escape>', self.cancel)
		
		self.parent.resizable(0, 0)
		self.parent.withdraw()
		self.parent.lift()
		self.text_entry.focus()
		self.parent.update_idletasks()
		x = self.host_app_parent.winfo_rootx() + self.host_app_parent.winfo_width() // 2 - self.parent.winfo_width() // 2
		y = self.host_app_parent.winfo_rooty() + self.host_app_parent.winfo_height() // 2 - self.parent.winfo_height() // 2
		self.parent.update_idletasks()
		self.parent.geometry('+%d+%d' % (x, y))
		self.parent.deiconify()
		self.parent.transient(self.host_app_parent)  # Lock this window on top.
		self.parent.grab_set()  # Make sure this window always has focus.
		self.parent.update_idletasks()
		self.host_app_parent.update_idletasks()
	
	def get_result(self):
		self.parent.wait_window(self.parent)  # Wait for the window to be killed.
		return self.result
	
	def ok(self, event=None):
		text = self.text_entry.get()
		if not self.validate or self.validate(text):
			self.result = text
			self.cancel()
		else:
			self.text_entry.selection_range(0, 'end')
	
	def cancel(self, event=None):
		self.host_app_parent.focus_set()
		self.parent.destroy()

def get_input(title, prompt, default='', validate=None, host_app_parent=None):
	return GetStringApp(title, prompt, default, validate, host_app_parent).get_result()

